# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gst, GstAudio, GstPbutils, GLib, Gio

from time import time
from gettext import gettext as _

from lollypop.tagreader import TagReader, Discoverer
from lollypop.player_plugins import PluginsPlayer
from lollypop.define import GstPlayFlags, App, StorageType, Repeat
from lollypop.codecs import Codecs
from lollypop.logger import Logger
from lollypop.objects_track import Track
from lollypop.utils import emit_signal, get_network_available


class BinPlayer:
    """
        Gstreamer bin player
    """

    def __init__(self):
        """
            Init playbin
        """
        # In the case of gapless playback, both 'about-to-finish'
        # and 'eos' can occur during the same stream.
        self.__track_in_pipe = False
        self.__cancellable = Gio.Cancellable()
        self.__codecs = Codecs()
        self._current_track = Track()
        self._next_track = Track()
        self._prev_track = Track()
        self._playbin = self._playbin1 = Gst.ElementFactory.make(
            "playbin", "player")
        self._playbin2 = Gst.ElementFactory.make("playbin", "player")
        self._plugins = self._plugins1 = PluginsPlayer(self._playbin1)
        self._plugins2 = PluginsPlayer(self._playbin2)
        for playbin in [self._playbin1, self._playbin2]:
            flags = playbin.get_property("flags")
            flags &= ~GstPlayFlags.GST_PLAY_FLAG_VIDEO
            playbin.set_property("flags", flags)
            playbin.set_property("buffer-size", 5 << 20)
            playbin.set_property("buffer-duration", 10 * Gst.SECOND)
            playbin.connect("notify::volume", self.__on_volume_changed)
            playbin.connect("about-to-finish",
                            self._on_stream_about_to_finish)
            bus = playbin.get_bus()
            bus.add_signal_watch()
            bus.connect("message::error", self._on_bus_error)
            bus.connect("message::eos", self._on_bus_eos)
            bus.connect("message::element", self._on_bus_element)
            bus.connect("message::stream-start", self._on_stream_start)
            bus.connect("message::tag", self._on_bus_message_tag)
        self._start_time = 0

    def load(self, track):
        """
            Load track and play it
            @param track as Track
        """
        self._playbin.set_state(Gst.State.NULL)
        if self._load_track(track):
            self.play()

    def play(self):
        """
            Change player state to PLAYING
        """
        # No current playback, song in queue
        if self._current_track.id is None:
            if self._next_track.id is not None:
                self.load(self._next_track)
        else:
            self._playbin.set_state(Gst.State.PLAYING)
            emit_signal(self, "status-changed")

    def pause(self):
        """
            Change player state to PAUSED
        """
        self._playbin.set_state(Gst.State.PAUSED)
        emit_signal(self, "status-changed")

    def stop(self):
        """
            Change player state to STOPPED
            @param force as bool
        """
        self._current_track = Track()
        self._current_track = Track()
        self._prev_track = Track()
        self._next_track = Track()
        emit_signal(self, "current-changed")
        emit_signal(self, "prev-changed")
        emit_signal(self, "next-changed")
        self._playbin.set_state(Gst.State.NULL)
        emit_signal(self, "status-changed")

    def stop_all(self):
        """
            Stop all bins, lollypop should quit now
        """
        # Stop
        self._playbin1.set_state(Gst.State.NULL)
        self._playbin2.set_state(Gst.State.NULL)

    def play_pause(self):
        """
            Set playing if paused
            Set paused if playing
        """
        if self.is_playing:
            self.pause()
        else:
            self.play()

    def reload_track(self):
        """
            Reload track at current position
        """
        if self.current_track.id is None:
            return
        position = self.position
        self.load(self.current_track)
        GLib.timeout_add(100, self.seek, position)

    def seek(self, position):
        """
            Seek current track to position
            @param position as int (ms)
        """
        if self._current_track.id is None:
            return
        # Seems gstreamer doesn't like seeking to end, sometimes
        # doesn't go to next track
        if position >= self._current_track.duration:
            self.next()
        else:
            self._playbin.seek_simple(Gst.Format.TIME,
                                      Gst.SeekFlags.FLUSH |
                                      Gst.SeekFlags.KEY_UNIT,
                                      position * 1000000)
            emit_signal(self, "seeked", position)

    def get_status(self):
        """
            Playback status
            @return Gstreamer state
        """
        ok, state, pending = self._playbin.get_state(Gst.CLOCK_TIME_NONE)
        if ok == Gst.StateChangeReturn.ASYNC:
            state = pending
        elif (ok != Gst.StateChangeReturn.SUCCESS):
            state = Gst.State.NULL
        return state

    def set_volume(self, rate):
        """
            Set player volume rate
            @param rate as double
        """
        if rate < 0.0:
            rate = 0.0
        elif rate > 1.0:
            rate = 1.0
        self._playbin.set_volume(GstAudio.StreamVolumeFormat.CUBIC, rate)

    @property
    def plugins(self):
        """
            Get plugins
            @return [PluginsPlayer]
        """
        return [self._plugins1, self._plugins2]

    @property
    def is_playing(self):
        """
            True if player is playing
            @return bool
        """
        ok, state, pending = self._playbin.get_state(Gst.CLOCK_TIME_NONE)
        if ok == Gst.StateChangeReturn.ASYNC:
            return pending == Gst.State.PLAYING
        elif ok == Gst.StateChangeReturn.SUCCESS:
            return state == Gst.State.PLAYING
        else:
            return False

    @property
    def position(self):
        """
            Return bin playback position
            @HACK handle crossefade here, as we know we're going to be
            called every seconds
            @return position as int (ms)
        """
        return self.__get_bin_position(self._playbin)

    @property
    def remaining(self):
        """
            Return remaining duration
            @return duration as int (ms)
        """
        position = self._playbin.query_position(Gst.Format.TIME)[1] / 1000000
        duration = self._current_track.duration
        return int(duration - position)

    @property
    def current_track(self):
        """
            Current track
        """
        return self._current_track

    @property
    def volume(self):
        """
            Return player volume rate
            @return rate as double
        """
        return self._playbin.get_volume(GstAudio.StreamVolumeFormat.CUBIC)

#######################
# PROTECTED           #
#######################
    def _load_track(self, track):
        """
            Load track
            @param track as Track
            @return False if track not loaded
        """
        self.__track_in_pipe = True
        Logger.debug("BinPlayer::_load_track(): %s" % track.uri)
        try:
            emit_signal(self, "loading-changed", False, self._current_track)
            self._current_track = track
            if track.is_web and not track.uri_loaded:
                emit_signal(self, "loading-changed", True, track)
                self.__load_from_web(track)
                return False
            else:
                self._playbin.set_property("uri", track.uri)
        except Exception as e:  # Gstreamer error
            Logger.error("BinPlayer::_load_track(): %s" % e)
            return False
        return True

    def _on_stream_start(self, bus, message):
        """
            On stream start
            Handle stream start: scrobbling, notify, ...
            @param bus as Gst.Bus
            @param message as Gst.Message
        """
        self.__track_in_pipe = False
        emit_signal(self, "loading-changed", False, self._current_track)
        self._start_time = time()
        Logger.debug("Player::_on_stream_start(): %s" %
                     self._current_track.uri)
        emit_signal(self, "current-changed")
        for scrobbler in App().ws_director.scrobblers:
            scrobbler.playing_now(self._current_track)

    def _on_bus_message_tag(self, bus, message):
        """
            Read tags from stream
            @param bus as Gst.Bus
            @param message as Gst.Message
        """
        if self._current_track.storage_type != StorageType.EXTERNAL:
            return
        Logger.debug("Player::__on_bus_message_tag(): %s" %
                     self._current_track.uri)
        reader = TagReader()
        tags = message.parse_tag()
        title = reader.get_title(tags, "")
        if len(title) > 1 and self._current_track.title != title:
            self._current_track.set_name(title)
            emit_signal(self, "current-changed")

    def _on_bus_element(self, bus, message):
        """
            Set elements for missings plugins
            @param bus as Gst.Bus
            @param message as Gst.Message
        """
        if GstPbutils.is_missing_plugin_message(message):
            self.__codecs.append(message)

    def _on_bus_error(self, bus, message):
        """
            Try a codec install and update current track
            @param bus as Gst.Bus
            @param message as Gst.Message
        """
        if self._current_track.is_web:
            emit_signal(self, "loading-changed", False,
                        self._current_track)
        Logger.info("Player::_on_bus_error(): %s" % message.parse_error()[1])
        if self.current_track.id is not None and self.current_track.id >= 0:
            if self.__codecs.is_missing_codec(message):
                self.__codecs.install()
                App().scanner.stop()
                self.stop()
            else:
                (error, parsed) = message.parse_error()
                App().notify.send("Lollypop", parsed)
                self.stop()

    def _on_bus_eos(self, bus, message):
        """
            If we are current bus, try to restart playback
            Else stop playback
        """
        if self.__track_in_pipe:
            return
        if self._playbin.get_bus() == bus:
            if self._next_track.id is None:
                self.stop()
            else:
                self._load_track(self._next_track)
                self.next()

    def _on_stream_about_to_finish(self, playbin):
        """
            When stream is about to finish, switch to next track without gap
            @param playbin as Gst.Bin
        """
        try:
            Logger.debug("Player::__on_stream_about_to_finish(): %s" % playbin)
            # Don't do anything if crossfade on, track already scrobbled
            # See TransitionsPlayer
            if not self.crossfading:
                self._on_track_finished(self.current_track)
                repeat = App().settings.get_enum("repeat")
                if repeat == Repeat.TRACK:
                    self._load_track(self.current_track)
                elif self._next_track.id is not None:
                    self._load_track(self._next_track)
        except Exception as e:
            Logger.error("BinPlayer::_on_stream_about_to_finish(): %s", e)

#######################
# PRIVATE             #
#######################
    def __load_from_web(self, track):
        """
            Load track from web
            @param track as Track
        """
        if get_network_available():
            self.__cancellable.cancel()
            self.__cancellable = Gio.Cancellable.new()
            from lollypop.helper_web import WebHelper
            helper = WebHelper(track, self.__cancellable)
            helper.connect("loaded", self.__on_web_helper_loaded,
                           track, self.__cancellable)
            helper.load()
        else:
            self.skip_album()

    def __get_bin_position(self, playbin):
        """
            Get position for playbin
            @param playbin as Gst.Bin
            @return position as int (ms)
        """
        return playbin.query_position(Gst.Format.TIME)[1] / 1000000

    def __update_current_duration(self, track):
        """
            Update current track duration
            @param track as Track
        """
        try:
            discoverer = Discoverer()
            duration = discoverer.get_info(track.uri).get_duration() / 1000000
            if duration != track.duration and duration > 0:
                App().tracks.set_duration(track.id, int(duration))
                track.reset("duration")
                emit_signal(self, "duration-changed", track.id)
        except Exception as e:
            Logger.error("BinPlayer::__update_current_duration(): %s" % e)

    def __on_volume_changed(self, playbin, sink):
        """
            Emit volume-changed signal
            @param playbin as Gst.Bin
            @param sink as Gst.Sink
        """
        App().settings.set_value("volume-rate", GLib.Variant("d", self.volume))
        emit_signal(self, "volume-changed")

    def __on_web_helper_loaded(self, helper, uri, track, cancellable):
        """
            Play track URI
            @param helper as WebHelper
            @param uri as str
            @param track as Track
            @param cancellable as Gio.Cancellable
        """
        if cancellable.is_cancelled():
            return
        if uri:
            track.set_uri(uri)
            track.set_preloaded()
            self.load(track)
            App().task_helper.run(self.__update_current_duration, track)
        else:
            GLib.idle_add(
                App().notify.send,
                "Lollypop",
                _("Can't find this track on YouTube"))
            self.next()



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GObject, GLib


from lollypop.widgets_tracks import TracksWidget
from lollypop.helper_signals import SignalsHelper, signals_map
from lollypop.define import App, ViewType, IndicatorType
from lollypop.define import Size
from lollypop.utils import emit_signal
from lollypop.helper_size_allocation import SizeAllocationHelper


class TracksView(Gtk.Bin, SignalsHelper, SizeAllocationHelper):
    """
        Responsive view showing tracks
    """

    __gsignals__ = {
        "activated": (GObject.SignalFlags.RUN_FIRST,
                      None, (GObject.TYPE_PYOBJECT,)),
        "selected": (GObject.SignalFlags.RUN_FIRST, None, (bool,)),
        "populated": (GObject.SignalFlags.RUN_FIRST, None, ()),
        "track-removed": (GObject.SignalFlags.RUN_FIRST, None,
                          (GObject.TYPE_PYOBJECT,)),
    }

    @signals_map
    def __init__(self, view_type):
        """
            Init view
            @param album as Album
            @param view_type as ViewType
        """
        Gtk.Bin.__init__(self)
        self._view_type = view_type
        self._tracks_widget_left = {}
        self._tracks_widget_right = {}
        self._responsive_widget = None
        self.__orientation = None
        self.connect("realize", self.__on_realize)
        return [
            (App().player, "loading-changed", "_on_loading_changed"),
            (App().player, "current-changed", "_on_current_changed"),
            (App().player, "duration-changed", "_on_duration_changed"),
        ]

    def get_current_ordinate(self, parent):
        """
            If current track in widget, return it ordinate,
            @param parent widget as Gtk.Widget
            @return y as int
        """
        for child in self.children:
            if child.id == App().player.current_track.id:
                return child.translate_coordinates(parent, 0, 0)[1]
        return None

    @property
    def children(self):
        """
            Return all rows
            @return [Gtk.ListBoxRow]
        """
        return []

    @property
    def boxes(self):
        """
            Get available list boxes
            @return [Gtk.ListBox]
        """
        return []

    @property
    def view_type(self):
        """
            Get view type
            @return ViewType
        """
        return self._view_type

#######################
# PROTECTED           #
#######################
    def _handle_width_allocate(self, allocation):
        """
            Change columns disposition
            @param allocation as Gtk.Allocation
        """
        if SizeAllocationHelper._handle_width_allocate(self, allocation):
            if allocation.width >= Size.NORMAL:
                orientation = Gtk.Orientation.HORIZONTAL
            else:
                orientation = Gtk.Orientation.VERTICAL
            if self.__orientation != orientation:
                self._set_orientation(orientation)

    def _init(self):
        """
            Init main widget
            @return bool
        """
        if self._responsive_widget is None:
            self._responsive_widget = Gtk.Grid()
            self._responsive_widget.set_column_spacing(20)
            self._responsive_widget.set_column_homogeneous(True)
            self._responsive_widget.set_property("valign", Gtk.Align.START)
            self.add(self._responsive_widget)
            self._responsive_widget.show()
            return True
        return False

    def _add_disc_container(self, disc_number):
        """
            Add disc container to box
            @param disc_number as int
        """
        self._tracks_widget_left[disc_number] = TracksWidget(self._view_type)
        self._tracks_widget_right[disc_number] = TracksWidget(self._view_type)
        self._tracks_widget_left[disc_number].connect(
            "activated", self._on_activated)
        self._tracks_widget_right[disc_number].connect(
            "activated", self._on_activated)
        self._tracks_widget_left[disc_number].connect(
            "do-selection", self.__on_do_selection)
        self._tracks_widget_right[disc_number].connect(
            "do-selection", self.__on_do_selection)
        self._tracks_widget_left[disc_number].connect(
            "do-shift-selection", self.__on_do_shift_selection)
        self._tracks_widget_right[disc_number].connect(
            "do-shift-selection", self.__on_do_shift_selection)
        self._tracks_widget_left[disc_number].connect(
            "row-selected", self._on_row_selected)
        self._tracks_widget_right[disc_number].connect(
            "row-selected", self._on_row_selected)

    def _add_tracks(self, widget, tracks, position=0):
        """
            Add tracks to widget
            @param widget as Gtk.ListBox
            @param tracks as [Track]
        """
        pass

    def _set_orientation(self, orientation):
        """
            Set columns orientation
            @param orientation as Gtk.Orientation
            @return updated as bool
        """
        if self.__orientation == orientation or\
                self._responsive_widget is None:
            return False
        self.__orientation = orientation
        for child in self._responsive_widget.get_children():
            self._responsive_widget.remove(child)
        return True

    def _allow_selection(self):
        """
            Allow selection on boxes
        """
        emit_signal(self, "selected", True)
        for box in self.boxes:
            if box.get_selection_mode() == Gtk.SelectionMode.MULTIPLE:
                continue
            box.set_selection_mode(Gtk.SelectionMode.MULTIPLE)

    def _disallow_selection(self):
        """
            Disallow selection on boxes
        """
        emit_signal(self, "selected", False)
        for box in self.boxes:
            if box.get_selection_mode() == Gtk.SelectionMode.NONE:
                continue
            box.set_selection_mode(Gtk.SelectionMode.NONE)
            box.set_activate_on_single_click(True)

    def _on_loading_changed(self, player, status, track):
        """
            Update row loading status
            @param player as Player
            @param status as bool
            @param track as Track
        """
        for row in self.children:
            if row.track.id == track.id:
                row.set_indicator(IndicatorType.LOADING)
            else:
                row.set_indicator()

    def _on_row_selected(self, listbox, row):
        """
            Update selection state
            @param listbox as Gtk.ListBox
            @param row as Gtk.ListBoxRow
        """
        for child in self.children:
            if child.is_selected():
                return
        self._disallow_selection()

    def _on_activated(self, widget, track):
        pass

    def _on_current_changed(self, player):
        """
            Update children state
            @param player as Player
        """
        for key in self._tracks_widget_left.keys():
            self._tracks_widget_left[key].update_playing(
                    App().player.current_track.id)
        for key in self._tracks_widget_right.keys():
            self._tracks_widget_right[key].update_playing(
                    App().player.current_track.id)

    def _on_duration_changed(self, player, track_id):
        """
            Update track duration
            @param player as Player
            @param track_id as int
        """
        for key in self._tracks_widget_left.keys():
            self._tracks_widget_left[key].update_duration(track_id)
        for key in self._tracks_widget_right.keys():
            self._tracks_widget_right[key].update_duration(track_id)

#######################
# PRIVATE             #
#######################
    def __on_realize(self, widget):
        """
            Set initial orientation
            @param widget as Gtk.Widget
            @param orientation as Gtk.Orientation
        """
        if self._view_type & ViewType.SINGLE_COLUMN or\
                App().settings.get_value("force-single-column"):
            self._set_orientation(Gtk.Orientation.VERTICAL)
        elif self._view_type & ViewType.TWO_COLUMNS:
            self._set_orientation(Gtk.Orientation.HORIZONTAL)
        else:
            # We need to listen to parent allocation as currently, we have
            # no children
            SizeAllocationHelper.__init__(self, True)

    def __on_do_selection(self, listbox, row):
        """
            Do shift selection
            @param listbox as Gtk.ListBox
            @param row as Gtk.ListBoxrow
        """
        if row is None:
            self._disallow_selection()
        else:
            self._allow_selection()
            if row.is_selected():
                # Let current selection event terminate
                GLib.timeout_add(100, listbox.unselect_row, row)
            else:
                listbox.select_row(row)

    def __on_do_shift_selection(self, listbox, row):
        """
            Do shift selection
            @param listbox as Gtk.ListBox
            @param row as Gtk.ListBoxrow
        """
        self._allow_selection()
        is_selected = row.is_selected()
        start_idx = end_idx = -1
        idx = 0
        for children in self.children:
            if children.is_selected() and not is_selected:
                start_idx = idx
            elif not children.is_selected() and is_selected:
                start_idx = idx
            if children == row:
                end_idx = idx
                break
            idx += 1
        if start_idx == -1:
            start_idx = 0
        for child in self.children[start_idx:end_idx + 1]:
            if is_selected:
                GLib.timeout_add(100,
                                 children.get_parent().unselect_row, child)
            else:
                children.get_parent().select_row(child)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from lollypop.objects_track import Track
from lollypop.utils import emit_signal


class QueuePlayer:
    """
        Manage queue
    """

    def __init__(self):
        """
            Init queue
        """
        self.__queue = []
        self._queue_current_track = None

    def set_queue(self, queue):
        """
            Set queue
            @param queue as [int]
        """
        self.__queue = queue

    def append_to_queue(self, track_id, notify=True):
        """
            Append track to queue,
            remove previous track if exist
            @param track_id as int
            @param notify as bool
        """
        if track_id in self.__queue:
            self.__queue.remove(track_id)
        self.__queue.append(track_id)
        self.set_next()
        self.set_prev()
        if notify:
            emit_signal(self, "queue-changed")

    def insert_in_queue(self, track_id, pos=0, notify=True):
        """
            Prepend track to queue,
            remove previous track if exist
            @param track_id as int
            @param pos as int
            @param notify as bool
        """
        if track_id in self.__queue:
            self.__queue.remove(track_id)
        self.__queue.insert(pos, track_id)
        self.set_next()
        self.set_prev()
        if notify:
            emit_signal(self, "queue-changed")

    def remove_from_queue(self, track_id, notify=True):
        """
            Remove track from queue
            @param track_id as int
            @param notify as bool
        """
        if track_id in self.__queue:
            self.__queue.remove(track_id)
        if notify:
            emit_signal(self, "queue-changed")

    def clear_queue(self, notify=True):
        """
            Set queue
            @param [ids as int]
            @param notify as bool
        """
        self.__queue = []
        if notify:
            emit_signal(self, "queue-changed")

    def is_in_queue(self, track_id):
        """
            True if track id exists in queue
            @param track_id as int
            @return bool
        """
        if self.__queue:
            return track_id in self.__queue
        else:
            return False

    def album_in_queue(self, album):
        """
            True if album id exists in queue
            @param album as Album
            @return bool
        """
        if self.__queue:
            union = set(self.__queue) & set(album.track_ids)
            return len(union) == len(album.track_ids)
        else:
            return False

    def get_track_position(self, track_id):
        """
            Return track position in queue
            @param track_id as int
            @return position as int
        """
        return self.__queue.index(track_id) + 1

    def next(self):
        """
            Get next track id
            @return Track
        """
        track_id = None
        if self.__queue:
            track_id = self.__queue[0]
            if self._queue_current_track is None:
                self._queue_current_track = self._current_track
        return Track(track_id)

    @property
    def queue(self):
        """
            Return queue
            @return [ids as int]
        """
        return self.__queue

#######################
# PRIVATE             #
#######################



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio

from gettext import gettext as _

from lollypop.define import App


class CurrentAlbumsMenu(Gio.Menu):
    """
        Menu for current albums
    """

    def __init__(self, header=False):
        """
            Init menu
            @param header as bool
        """
        Gio.Menu.__init__(self)

        if header:
            from lollypop.menu_header import MenuHeader
            label = _("Playing albums")
            icon_name = "org.gnome.Lollypop-play-queue-symbolic"
            self.append_item(MenuHeader(label, icon_name))
        menu = Gio.Menu()
        save_playback_action = Gio.SimpleAction.new(name="save_playback")
        save_playback_action.connect(
            "activate", self.__on_save_playback_action_activate)
        App().add_action(save_playback_action)
        menu.append(_("Create a new playlist"), "app.save_playback")
        self.append_section(_("Playing albums"), menu)

    def __on_save_playback_action_activate(self, action, variant):
        """
            Create a new playlist based on current playback
            @param Gio.SimpleAction
            @param GLib.Variant
        """
        def albums_to_playlist():
            tracks = []
            for album in App().player.albums:
                tracks += album.tracks
            if tracks:
                import datetime
                now = datetime.datetime.now()
                date_string = now.strftime("%Y-%m-%d-%H:%M:%S")
                playlist_id = App().playlists.add(date_string)
                App().playlists.add_tracks(playlist_id, tracks)
        App().task_helper.run(albums_to_playlist)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GObject

from lollypop.define import App, ViewType, MARGIN
from lollypop.view_tracks_album import AlbumTracksView
from lollypop.widgets_banner_album import AlbumBannerWidget
from lollypop.utils import emit_signal
from lollypop.helper_gestures import GesturesHelper


class AlbumWidget(Gtk.Grid):
    """
        Show artist albums and tracks
    """

    __gsignals__ = {
        "populated": (GObject.SignalFlags.RUN_FIRST, None, ()),
    }

    def __init__(self, album, storage_type, view_type):
        """
            Init album widget
            @param album as Album
            @param storage_type as StorageType
            @param view_type as ViewType
        """
        Gtk.Grid.__init__(self)
        self.set_orientation(Gtk.Orientation.VERTICAL)
        self.__tracks_view = None
        self.__revealer = None
        self.__view_type = view_type
        self.__storage_type = storage_type
        self.__album = album
        if App().settings.get_value("force-single-column"):
            self.__column_width = 1200
        else:
            self.__column_width = 600

    def populate(self):
        """
            Populate widget
        """
        if self.__revealer is None:
            self.__revealer = Gtk.Revealer.new()
            self.__revealer.show()
            self.__banner = AlbumBannerWidget(self.__album,
                                              self.__storage_type,
                                              self.__view_type)
            self.__banner.show()
            self.__banner.connect("populated", self.__on_banner_populated)
            self.__banner.populate()
            self.add(self.__banner)
            self.add(self.__revealer)
            self.__gesture = GesturesHelper(
                self.__banner,  primary_press_callback=self._on_press)
            self.get_style_context().add_class("album-banner")
            if App().settings.get_value("show-artist-tracks"):
                self.__revealer.set_transition_type(
                    Gtk.RevealerTransitionType.NONE)
                self.__populate()
            self.set_selection()

    def reveal_child(self):
        """
            Reveal tracks
        """
        self.__revealer.set_transition_type(
                Gtk.RevealerTransitionType.NONE)
        self.__populate()

    def set_selection(self):
        """
            Hilight widget if currently playing
        """
        selected = self.__album.id == App().player.current_track.album.id
        if selected:
            self.set_state_flags(Gtk.StateFlags.SELECTED, False)
        else:
            self.unset_state_flags(Gtk.StateFlags.SELECTED)

    def set_duration(self, track_id):
        """
            Update track duration
            @param track_id as int
        """
        if self.__tracks_view is not None:
            self.__tracks_view.update_duration(track_id)

    def do_get_preferred_width(self):
        return (200, self.__column_width)

    @property
    def banner(self):
        """
            Get banner
            @return BannerWidget
        """
        if self.__revealer is None:
            self.populate()
        return self.__banner

    @property
    def name(self):
        """
            Get name
            @return str
        """
        return self.__album.name

    @property
    def album(self):
        """
            Get album
            @return Album
        """
        return self.__album

    @property
    def is_populated(self):
        """
            True if populated
            @return bool
        """
        return True

    @property
    def args(self):
        """
            Get default args for __class__
            @return {}
        """
        return {"album": self.__album,
                "storage_type": self.storage_type,
                "view_type": self.view_type & ~ViewType.SMALL}

    @property
    def filtered(self):
        """
            Get filtered children
            @return [Gtk.Widget]
        """
        filtered = []
        if self.__tracks_view is None:
            self.__populate()
        filtered = self.__tracks_view.children
        return filtered

    @property
    def scroll_shift(self):
        """
            Get scroll shift for y axes
            @return int
        """
        return self.__banner.height

#######################
# PROTECTED           #
#######################
    def _on_press(self, x, y, event):
        """
            Show tracks
            @param x as int
            @param y as int
            @param event as Gdk.Event
        """
        self.__populate()

#######################
# PRIVATE             #
#######################
    def __populate(self):
        """
            Populate the view with album
        """
        def init_tracks_view():
            if self.__tracks_view is None:
                self.__tracks_view = AlbumTracksView(self.__album,
                                                     self.__view_type)
                self.__tracks_view.show()
                self.__tracks_view.set_margin_start(MARGIN)
                self.__tracks_view.set_margin_end(MARGIN)
                self.__tracks_view.populate()
                self.__tracks_view.connect("populated",
                                           self.__on_tracks_populated)
                self.__tracks_view.connect("selected",
                                           self.__on_track_selected)
                self.__revealer.add(self.__tracks_view)

        def reveal(ignore, update_duration):
            if update_duration:
                self.get_style_context().remove_class("load-animation")
                self.__banner.update_duration()
            self.__revealer.set_reveal_child(
                not self.__revealer.get_reveal_child())
        init_tracks_view()
        if self.__album.tracks:
            reveal(None, False)
        else:
            self.get_style_context().add_class("load-animation")
            App().task_helper.run(self.__album.load_tracks,
                                  None,
                                  callback=(reveal, True))

    def __on_tracks_populated(self, view):
        """
            Populate remaining discs
            @param view as TracksView
        """
        if not self.__tracks_view.is_populated:
            self.__tracks_view.populate()
        else:
            self.__revealer.set_transition_type(
                Gtk.RevealerTransitionType.SLIDE_DOWN)

    def __on_track_selected(self, view, selected):
        """
            Mark banner as filtered
            @param view as TracksView
            @param selected as bool
        """
        self.__banner.set_selected(selected)

    def __on_banner_populated(self, widget):
        """
            Emit populated signal
            @param widget as Gtk.Widget
        """
        emit_signal(self, "populated")



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib

import json

from lollypop.define import App, AUDIODB_CLIENT_ID
from lollypop.define import FANARTTV_ID
from lollypop.define import StorageType
from lollypop.utils import get_network_available, emit_signal
from lollypop.logger import Logger
from lollypop.artwork_downloader import ArtworkDownloader


class ArtistArtworkDownloader(ArtworkDownloader):
    """
        Download artwork from the web
    """

    def __init__(self):
        """
            Init artwork downloader
        """
        ArtworkDownloader.__init__(self)
        self.__methods = {
            "AudioDB": self.__get_audiodb_artist_artwork_uri,
            "FanartTV": self.__get_fanarttv_artist_artwork_uri,
            "Spotify": self.__get_spotify_artist_artwork_uri,
            "Deezer": self.__get_deezer_artist_artwork_uri
        }
        self.__queue = []
        self.__downloading = False

    def add_from_uri(self, artist, uri, cancellable, storage_type):
        """
            Add artist artwork from URI
            @param artist as str
            @param uri as str
            @param cancellable as Gio.Cancellable
            @param storage_type as int
        """
        def on_uri_content(uri, status, data):
            if status:
                self.add(artist, data, storage_type)
                emit_signal(self, "artist-artwork-changed", artist)
        if uri:
            App().task_helper.load_uri_content(uri,
                                               cancellable,
                                               on_uri_content)

    def download(self, artist):
        """
            Cache artist artwork
            @param artist as str
        """
        if not get_network_available("DATA"):
            return
        self.__queue.append(artist)
        if not self.__downloading:
            App().task_helper.run(self.__download_queue)

    def search(self, artist, cancellable):
        """
            Search artist artwork
            @param album as str
            @param cancellable as Gio.Cancellable
        """
        results = []
        for api in self.__methods.keys():
            if cancellable.is_cancelled():
                return
            uris = self.__methods[api](artist, cancellable)
            for uri in uris:
                if cancellable.is_cancelled():
                    return
                results.append((uri, api))
        emit_signal(self, "uri-artwork-found", results)

#######################
# PRIVATE             #
#######################
    def __get_audiodb_artist_artwork_uri(self, artist, cancellable=None):
        """
            Get artist artwork using AutdioDB
            @param artist as str
            @param cancellable as Gio.Cancellable
            @return uris as [str]
        """
        if not get_network_available("AUDIODB"):
            return []
        uris = []
        try:
            artist = GLib.uri_escape_string(artist, None, True)
            uri = "https://theaudiodb.com/api/v1/json/"
            uri += "%s/search.php?s=%s" % (AUDIODB_CLIENT_ID, artist)
            (status, data) = App().task_helper.load_uri_content_sync(
                uri, cancellable)
            if status:
                decode = json.loads(data.decode("utf-8"))
                for item in decode["artists"]:
                    for key in ["strArtistFanart", "strArtistThumb"]:
                        uri = item[key]
                        if uri is not None:
                            uris.append(uri)
        except:
            Logger.error("AudioDB: %s", uri)
        return uris

    def __get_deezer_artist_artwork_uri(self, artist, cancellable=None):
        """
            Get artist artwork using Deezer
            @param artist as str
            @param cancellable as Gio.Cancellable
            @return uris as [str]
        """
        if not get_network_available("DEEZER"):
            return []
        uris = []
        try:
            artist_formated = GLib.uri_escape_string(
                artist, None, True).replace(" ", "+")
            uri = "https://api.deezer.com/search/artist/?" +\
                  "q=%s&output=json&index=0" % artist_formated
            (status, data) = App().task_helper.load_uri_content_sync(
                uri, cancellable)
            if status:
                decode = json.loads(data.decode("utf-8"))
                for item in decode["data"]:
                    uri = item["picture_xl"]
                    uris.append(uri)
        except:
            Logger.error("Deezer: %s", uri)
        return uris

    def __get_fanarttv_artist_artwork_uri(self, artist, cancellable=None):
        """
            Get artist artwork using FanartTV
            @param artist as str
            @param cancellable as Gio.Cancellable
            @return uris as [str]
        """
        if not get_network_available("FANARTTV"):
            return []
        uris = []
        try:
            mbid = self._get_musicbrainz_mbid("artist", artist, cancellable)
            if mbid is None:
                return []
            uri = "http://webservice.fanart.tv/v3/music/%s?api_key=%s" % (
                mbid, FANARTTV_ID)
            (status, data) = App().task_helper.load_uri_content_sync(
                uri, cancellable)
            if status:
                decode = json.loads(data.decode("utf-8"))
                for item in decode["artistbackground"]:
                    uris.append(item["url"])
        except:
            Logger.error("FanartTV: %s", uri)
        return uris

    def __get_spotify_artist_artwork_uri(self, artist, cancellable=None):
        """
            Get artist artwork using Spotify
            @param artist as str
            @param cancellable as Gio.Cancellable
            @return uris as [str]
        """
        if not get_network_available("SPOTIFY"):
            return []
        uris = []
        try:
            artist_formated = GLib.uri_escape_string(
                artist, None, True).replace(" ", "+")
            uri = "https://api.spotify.com/v1/search?q=%s" % artist_formated +\
                  "&type=artist"
            token = App().ws_director.token_ws.get_token("SPOTIFY",
                                                         cancellable)
            bearer = "Bearer %s" % token
            headers = [("Authorization", bearer)]
            (status,
             data) = App().task_helper.load_uri_content_sync_with_headers(
                    uri, headers, cancellable)
            if status:
                decode = json.loads(data.decode("utf-8"))
                for item in decode["artists"]["items"]:
                    image_uri = item["images"][0]["url"]
                    uris.append(image_uri)
        except:
            Logger.error("Spotify: %s", uri)
        return uris

    def __download_queue(self):
        """
            Cache artwork for all artists
        """
        self.__downloading = True
        try:
            while self.__queue:
                artist = self.__queue.pop()
                found = False
                for api in self.__methods.keys():
                    result = self.__methods[api](artist)
                    for uri in result:
                        found = True
                        self.add_from_uri(artist, uri, self.cancellable,
                                          StorageType.COLLECTION)
                        break
                    # Found, do not search in another helper
                    if found:
                        break
                # Not found, save empty artwork
                if not found:
                    self.add(artist, None, StorageType.COLLECTION)
        except Exception as e:
            Logger.error("ArtistArtworkDownloader::__download_queue(): %s" % e)
        self.__downloading = False



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, Gio, GLib, Pango

from gettext import gettext as _

from lollypop.define import App, ArtSize, ArtBehaviour, Type, StorageType
from lollypop.utils import sql_escape
from lollypop.utils_file import get_youtube_dl
from lollypop.helper_signals import SignalsHelper, signals_map


class ArtistRow(Gtk.ListBoxRow):
    """
        An artist row
    """

    def __init__(self, artist_name, cover_uri, cancellable, storage_type):
        """
            Init row
            @param artist_name as str
            @param cover_uri as str
            @param cancellable as Gio.Cancellable
            @param storage_type as StorageType
        """
        Gtk.ListBoxRow.__init__(self)
        self.__artist_name = artist_name
        self.__cover_uri = cover_uri
        self.__cancellable = cancellable
        self.__storage_type = storage_type
        grid = Gtk.Grid()
        grid.set_column_spacing(5)
        label = Gtk.Label.new(artist_name)
        label.set_property("halign", Gtk.Align.START)
        label.set_ellipsize(Pango.EllipsizeMode.END)
        self.__artwork = Gtk.Image.new()
        self.__artwork.set_size_request(ArtSize.SMALL,
                                        ArtSize.SMALL)
        self.set_artwork()
        grid.add(self.__artwork)
        grid.add(label)
        grid.show_all()
        self.add(grid)

    def set_artwork(self):
        """
            Set artist artwork
        """
        if App().artist_art.get_path(self.__artist_name) is not None:
            App().art_helper.set_artist_artwork(self.__artist_name,
                                                ArtSize.SMALL,
                                                ArtSize.SMALL,
                                                self.get_scale_factor(),
                                                ArtBehaviour.CROP |
                                                ArtBehaviour.CACHE |
                                                ArtBehaviour.ROUNDED,
                                                self.__on_artist_artwork)
        else:
            self.__on_artist_artwork(None)

    @property
    def artist_name(self):
        """
            Get artist name
            @return str
        """
        return self.__artist_name

    @property
    def storage_type(self):
        """
            Get storage type
            @param storage type as StorageType
        """
        return self.__storage_type

#######################
# PRIVATE             #
#######################
    def __on_artist_artwork(self, surface):
        """
            Set artist artwork
            @param surface as cairo.Surface
        """
        if surface is None:
            # Last chance to get a cover
            if self.__cover_uri is not None:
                App().artist_art.add_from_uri(self.__artist_name,
                                              self.__cover_uri,
                                              self.__cancellable,
                                              StorageType.EPHEMERAL)
                self.__cover_uri = None
            # Cache for later usage
            else:
                App().artist_art.download(self.__artist_name)
            self.__artwork.get_style_context().add_class("circle-icon")
            self.__artwork.set_from_icon_name("avatar-default-symbolic",
                                              Gtk.IconSize.INVALID)
            # circle-icon padding is 5px
            self.__artwork.set_pixel_size(ArtSize.SMALL - 20)
        else:
            self.__artwork.get_style_context().remove_class("circle-icon")
            self.__artwork.set_from_surface(surface)
            del surface


class SimilarsMenu(Gtk.Bin, SignalsHelper):
    """
        A popover with similar artists
    """

    @signals_map
    def __init__(self, artist_id):
        """
            Init popover
            @param artist_id as int
        """
        Gtk.Bin.__init__(self)
        (path, env) = get_youtube_dl()
        self.__show_all = path is not None
        self.__added = []
        self.__artist_id = artist_id
        self.__cancellable = Gio.Cancellable()
        self.connect("map", self.__on_map)
        self.connect("unmap", self.__on_unmap)
        self.__stack = Gtk.Stack.new()
        self.__stack.show()
        self.__stack.set_transition_type(Gtk.StackTransitionType.CROSSFADE)
        self.__stack.set_transition_duration(200)
        self.__label = Gtk.Label.new(_("Loading…"))
        self.__label.get_style_context().add_class("bold")
        self.__label.get_style_context().add_class("dim-label")
        self.__label.show()
        self.__listbox = Gtk.ListBox()
        self.__listbox.get_style_context().add_class("trackswidget")
        self.__listbox.set_vexpand(True)
        self.__listbox.set_selection_mode(Gtk.SelectionMode.NONE)
        self.__listbox.set_activate_on_single_click(True)
        self.__listbox.connect("row-activated", self.__on_row_activated)
        self.__listbox.set_sort_func(self.__sort_func)
        self.__listbox.show()
        self.__stack.add(self.__label)
        self.__stack.add(self.__listbox)
        self.add(self.__stack)
        self.set_hexpand(True)
        return [
            (App().artist_art, "artist-artwork-changed",
             "_on_artist_artwork_changed")
        ]

    @property
    def section(self):
        """
            Get section
            @return GLib.Variant
        """
        return GLib.Variant("s", _("Others"))

    @property
    def submenu_name(self):
        """
            Get submenu name
            @return str
        """
        return _("Similar artists")

#######################
# PROTECTED           #
#######################
    def _on_artist_artwork_changed(self, art, artist_name):
        """
            Reload artist artwork
            @param art as Art
            @param artist_name as str
        """
        for child in self.__listbox.get_children():
            if child.artist_name == artist_name:
                child.set_artwork()
                break

#######################
# PRIVATE             #
#######################
    def __sort_func(self, row_a, row_b):
        """
            Sort rows
            @param row_a as Gtk.ListBoxRow
            @param row_b as Gtk.ListBoxRow
        """
        if row_a.storage_type == StorageType.COLLECTION and\
                row_b.storage_type == StorageType.EPHEMERAL:
            return False
        elif row_a.storage_type == StorageType.EPHEMERAL and\
                row_b.storage_type == StorageType.COLLECTION:
            return True
        else:
            return False

    def __on_map(self, widget):
        """
            Populate widget
            @param widget as Gtk.Widget
        """
        if self.__added:
            return
        from lollypop.similars import Similars
        similars = Similars()
        App().task_helper.run(
            similars.get_similar_artists,
            [self.__artist_id],
            self.__cancellable,
            callback=(self.__on_get_similar_artists,))

    def __on_unmap(self, widget):
        """
            Cancel loading
            @param widget as Gtk.Widget
        """
        self.__cancellable.cancel()

    def __on_row_activated(self, widget, row):
        """
            Play searched item when selected
            @param widget as Gtk.ListBox
            @param row as Gtk.ListBoxRow
        """
        # Close popover
        popover = self.get_ancestor(Gtk.Popover)
        if popover is not None:
            popover.hide()
        artist_name = row.artist_name
        if row.storage_type == StorageType.EPHEMERAL:
            App().lookup_action("search").activate(
                GLib.Variant("s", artist_name))
        else:
            artist_id = App().artists.get_id_for_escaped_string(
                sql_escape(artist_name))
            App().window.container.show_view([Type.ARTISTS], [artist_id])

    def __on_get_similar_artists(self, artists):
        """
            Add artist to view
            @param artists as [(str, str)]
        """
        if artists:
            self.__stack.set_visible_child(self.__listbox)
            self.__label.hide()
            (artist, cover_uri) = artists.pop(0)
            if artist in self.__added:
                GLib.idle_add(self.__on_get_similar_artists, artists)
                return
            self.__added.append(artist)
            artist_id = App().artists.get_id_for_escaped_string(
                sql_escape(artist))
            row = None
            if artist_id is not None and App().artists.has_albums(artist_id):
                # We want real artist name (with case)
                artist = App().artists.get_name(artist_id)
                row = ArtistRow(artist, None, self.__cancellable,
                                StorageType.COLLECTION)
            if row is None and self.__show_all:
                row = ArtistRow(artist, cover_uri, self.__cancellable,
                                StorageType.EPHEMERAL)
            if row is not None:
                row.show()
                self.__listbox.add(row)
            GLib.idle_add(self.__on_get_similar_artists, artists)
        else:
            self.__label.set_text(_("No results"))



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk

from random import shuffle

from lollypop.define import App, ArtBehaviour, StorageType
from lollypop.utils import emit_signal
from lollypop.objects_album import Album
from lollypop.widgets_flowbox_rounded import RoundedFlowBoxWidget


class RoundedArtistWidget(RoundedFlowBoxWidget):
    """
        Artist photo or artist's albums in a rounded widget
    """

    def __init__(self, item, view_type, font_height):
        """
            Init widget
            @param item as (int, str, str)
            @param view_type as ViewType
            @param font_height as int
        """
        # Get values from DB
        if isinstance(item, int):
            artist_id = item
            artist_name = artist_sortname = App().artists.get_name(artist_id)
        else:
            artist_id = item[0]
            artist_name = item[1]
            artist_sortname = item[2]
        RoundedFlowBoxWidget.__init__(self, artist_id, artist_name,
                                      artist_sortname, view_type, font_height)

    def populate(self):
        """
            Populate widget content
        """
        if self._artwork is None:
            RoundedFlowBoxWidget.populate(self)
            self._artwork.get_style_context().add_class("circle-icon-large")
            self.connect("destroy", self.__on_destroy)
        else:
            self.set_artwork()

    def set_artwork(self):
        """
            Set artist artwork
        """
        self._set_artwork()

#######################
# PROTECTED           #
#######################
    def _set_artwork(self):
        """
            Set artist artwork
        """
        if self._artwork is None:
            return
        RoundedFlowBoxWidget.set_artwork(self)
        if App().settings.get_value("artist-artwork"):
            App().art_helper.set_artist_artwork(
                                            self.name,
                                            self._art_size,
                                            self._art_size,
                                            self._artwork.get_scale_factor(),
                                            ArtBehaviour.ROUNDED |
                                            ArtBehaviour.CROP_SQUARE |
                                            ArtBehaviour.CACHE,
                                            self.__on_artist_artwork)
        else:
            album_ids = App().albums.get_ids([], [self._data],
                                             StorageType.ALL, True)
            if album_ids:
                shuffle(album_ids)
                App().art_helper.set_album_artwork(
                                            Album(album_ids[0]),
                                            self._art_size,
                                            self._art_size,
                                            self._artwork.get_scale_factor(),
                                            ArtBehaviour.ROUNDED |
                                            ArtBehaviour.CROP_SQUARE |
                                            ArtBehaviour.CACHE,
                                            self.__on_artist_artwork)
            else:
                self.__on_artist_artwork(None)

#######################
# PRIVATE             #
#######################
    def __on_artist_artwork(self, surface):
        """
            Finish widget initialisation
            @param surface as cairo.Surface
        """
        if self._artwork is None:
            return
        if surface is None:
            self._artwork.set_from_icon_name("avatar-default-symbolic",
                                             Gtk.IconSize.DIALOG)
        else:
            self._artwork.set_from_surface(surface)
        emit_signal(self, "populated")

    def __on_destroy(self, widget):
        """
            Destroyed widget
            @param widget as Gtk.Widget
        """
        self.__artwork = None



# Copyright (c) 2014-2017 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio

from time import time

from lollypop.logger import Logger
from lollypop.sqlcursor import SqlCursor
from lollypop.ws_collection_spotify import SpotifyCollectionWebService
from lollypop.ws_collection_deezer import DeezerCollectionWebService
from lollypop.helper_web_save import SaveWebHelper
from lollypop.define import App, StorageType, NetworkAccessACL


class CollectionWebService(SaveWebHelper,
                           SpotifyCollectionWebService,
                           DeezerCollectionWebService):
    """
        Add items to collection based on current user settings
    """
    MIN_ITEMS_PER_STORAGE_TYPE = 10
    MAX_ITEMS_PER_STORAGE_TYPE = 50
    __METHODS = {
        StorageType.SPOTIFY_SIMILARS:
            SpotifyCollectionWebService.search_similar_albums,
        StorageType.SPOTIFY_NEW_RELEASES:
            SpotifyCollectionWebService.search_new_releases,
        StorageType.DEEZER_CHARTS:
            DeezerCollectionWebService.search_charts
    }

    def __init__(self):
        """
            Init object
        """
        SaveWebHelper.__init__(self)
        SpotifyCollectionWebService.__init__(self)
        DeezerCollectionWebService.__init__(self)
        self.__is_running = False
        self.__cancellable = Gio.Cancellable()

    def start(self):
        """
            Populate DB in a background task
        """
        if self.__is_running:
            return
        App().task_helper.run(self.__populate_db)
        return True

    def stop(self):
        """
            Stop db populate
            @return bool
        """
        self.__cancellable.cancel()
        return not self.__is_running

#######################
# PRIVATE             #
#######################
    def __populate_db(self):
        """
            Populate DB in a background task
        """
        try:
            monitor = Gio.NetworkMonitor.get_default()
            if monitor.get_network_metered():
                return
            Logger.info("Collection download started")
            self.__is_running = True
            self.__cancellable = Gio.Cancellable()
            acl_storage_types = []
            acl = App().settings.get_value("network-access-acl").get_int32()
            mask = App().settings.get_value("suggestions-mask").get_int32()
            if acl & NetworkAccessACL["SPOTIFY"]:
                acl_storage_types += [StorageType.SPOTIFY_SIMILARS,
                                      StorageType.SPOTIFY_NEW_RELEASES]
            if acl & NetworkAccessACL["DEEZER"]:
                acl_storage_types.append(StorageType.DEEZER_CHARTS)
            # Check if storage type needs to be updated
            # Check if albums newer than a week are enough
            timestamp = time() - 604800
            storage_types = []
            for storage_type in acl_storage_types:
                if not mask & storage_type:
                    continue
                newer_albums = App().albums.get_newer_for_storage_type(
                                                           storage_type,
                                                           timestamp)
                if len(newer_albums) < self.MIN_ITEMS_PER_STORAGE_TYPE:
                    storage_types.append(storage_type)
            # Update needed storage types
            if storage_types:
                for storage_type in storage_types:
                    if self.__cancellable.is_cancelled():
                        raise Exception("cancelled")
                    self.__METHODS[storage_type](self, self.__cancellable)
                self.clean_old_albums(storage_types)
                App().artists.update_featuring()
        except Exception as e:
            Logger.warning("CollectionWebService::__populate_db(): %s", e)
        self.__is_running = False
        Logger.info("Collection download finished")

    def clean_old_albums(self, storage_types):
        """
            Clean old albums from DB
            @param storage_types as [StorageType]
        """
        SqlCursor.add(App().db)
        # Remove older albums
        for storage_type in storage_types:
            # If too many albums, do some cleanup
            count = App().albums.get_count_for_storage_type(storage_type)
            diff = count - self.MAX_ITEMS_PER_STORAGE_TYPE
            if diff > 0:
                album_ids = App().albums.get_oldest_for_storage_type(
                    storage_type, diff)
                for album_id in album_ids:
                    # EPHEMERAL with not tracks will be cleaned below
                    App().albums.set_storage_type(album_id,
                                                  StorageType.EPHEMERAL)
                    App().tracks.remove_album(album_id, False)
        # On cancel, clean not needed, done in Application::quit()
        if not self.__cancellable.is_cancelled():
            App().tracks.clean(False)
            App().albums.clean(False)
            App().artists.clean(False)
        SqlCursor.remove(App().db)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from random import shuffle

from lollypop.view_flowbox import FlowBoxView
from lollypop.widgets_albums_decade import AlbumsDecadeWidget
from lollypop.define import App, Type, ViewType, OrderBy
from lollypop.utils import get_icon_name
from lollypop.objects_album import Album


class DecadesBoxView(FlowBoxView):
    """
        Show decades in a FlowBox
    """

    def __init__(self, storage_type, view_type):
        """
            Init decade view
            @param storage_type as StorageType
            @param view_type as ViewType
        """
        from lollypop.widgets_banner_flowbox import FlowboxBannerWidget
        FlowBoxView.__init__(self, storage_type, view_type | ViewType.OVERLAY)
        self._empty_icon_name = get_icon_name(Type.YEARS)
        self.__banner = FlowboxBannerWidget([Type.YEARS], [], self.view_type)
        self.__banner.show()
        self.__banner.connect("play-all", self.__on_banner_play_all)
        self.add_widget(self._box, self.__banner)

    def populate(self):
        """
            Populate view
        """
        def on_load(items):
            FlowBoxView.populate(self, items)

        def load():
            (years, unknown) = App().tracks.get_years(self.storage_type)
            decades = []
            decade = []
            current_d = None
            for year in sorted(years):
                d = year // 10
                if current_d is not None and current_d != d:
                    current_d = d
                    decades.append(decade)
                    decade = []
                current_d = d
                decade.append(year)
            if decade:
                decades.append(decade)
            return decades

        App().task_helper.run(load, callback=(on_load,))

    @property
    def args(self):
        """
            Get default args for __class__
            @return {}
        """
        return {"storage_type": self.storage_type,
                "view_type": self.view_type}

#######################
# PROTECTED           #
#######################
    def _get_child(self, value):
        """
            Get a child for view
            @param value as object
            @return row as SelectionListRow
        """
        if self.destroyed:
            return None
        widget = AlbumsDecadeWidget(value, self.view_type, self.font_height)
        self._box.insert(widget, -1)
        widget.show()
        return widget

    def _get_menu_widget(self, child):
        """
            Get menu widget
            @param child as AlbumSimpleWidget
            @return Gtk.Widget
        """
        from lollypop.widgets_menu import MenuBuilder
        from lollypop.menu_decade import DecadeMenu
        menu = DecadeMenu(child.data, self.view_type,
                          App().window.folded)
        return MenuBuilder(menu)

    def _on_child_activated(self, flowbox, child):
        """
            Navigate into child
            @param flowbox as Gtk.FlowBox
            @param child as Gtk.FlowBoxChild
        """
        App().window.container.show_view([Type.YEARS], child.data)

    def _on_tertiary_press_gesture(self, x, y, event):
        """
            Play artist
            @param x as int
            @param y as int
            @param event as Gdk.Event
        """
        child = self._box.get_child_at_pos(x, y)
        if child is None or child.artwork is None:
            return
        items = []
        for year in child.data:
            items += App().tracks.get_albums_by_disc_for_year(
                year, self.storage_type, False)
            items += App().tracks.get_compilations_by_disc_for_year(
                year, self.storage_type, False)
        album_ids = [item[0] for item in items]
        albums = [Album(album_id, [], [], False) for album_id in album_ids]
        if albums:
            App().player.play_album_for_albums(albums[0], albums)

#######################
# PRIVATE             #
#######################
    def __on_banner_play_all(self, banner, random):
        """
            Play all albums
            @param banner as AlbumsBannerWidget
            @param random as bool
        """
        album_ids = App().albums.get_ids([], [], self.storage_type,
                                         False, OrderBy.YEAR_ASC)
        if not album_ids:
            return
        albums = [Album(album_id, [], [], False) for album_id in album_ids]
        if random:
            shuffle(albums)
            App().player.play_album_for_albums(albums[0], albums)
        else:
            App().player.play_album_for_albums(albums[0], albums)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gio, GObject

from hashlib import md5

from lollypop.logger import Logger
from lollypop.information_downloader import InformationDownloader


class InformationStore(GObject.Object, InformationDownloader):
    """
        Generic class to cache information
    """

    def __init__(self):
        """
            Init store
        """
        GObject.Object.__init__(self)
        InformationDownloader.__init__(self)

    def get_information(self, name, path):
        """
            Get information for name and path
            @param name as str
            @param path as str
            @return content as bytes
        """
        content = None
        try:
            encoded = md5(name.encode("utf-8")).hexdigest()
            filepath = "%s/%s.txt" % (path, encoded)
            f = Gio.File.new_for_path(filepath)
            if f.query_exists():
                (status, content, tag) = f.load_contents()
        except Exception as e:
            Logger.error("InformationStore::get_information(): %s", e)
        return content

    def save_information(self, name, path, content):
        """
            Save information for name and path
            @param name as str
            @param path as str
            @param content as bytes
        """
        try:
            if content is not None:
                encoded = md5(name.encode("utf-8")).hexdigest()
                filepath = "%s/%s.txt" % (path, encoded)
                f = Gio.File.new_for_path(filepath)
                fstream = f.replace(None, False,
                                    Gio.FileCreateFlags.REPLACE_DESTINATION,
                                    None)
                if fstream is not None:
                    fstream.write(content, None)
                    fstream.close()
        except Exception as e:
            Logger.error("InformationStore::save_information(): %s", e)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from random import choice, shuffle
import json

from lollypop.logger import Logger
from lollypop.define import App
from lollypop.utils import emit_signal
from lollypop.helper_web_spotify import SpotifyWebHelper
from lollypop.helper_web_save import SaveWebHelper


class SpotifySimilars(SaveWebHelper, SpotifyWebHelper):
    """
        Search similar artists with Spotify
    """
    def __init__(self):
        """
            Init provider
        """
        SaveWebHelper.__init__(self)
        SpotifyWebHelper.__init__(self)

    def load_similars(self, artist_ids, storage_type, cancellable):
        """
            Load similar artists for artist ids
            @param artist_ids as int
            @param storage_type as StorageType
            @param cancellable as Gio.Cancellable
        """
        names = [App().artists.get_name(artist_id) for artist_id in artist_ids]
        spotify_ids = self.get_similar_artist_ids(names, cancellable)
        track_ids = []
        for spotify_id in spotify_ids:
            spotify_ids = self.get_artist_top_tracks(spotify_id, cancellable)
            # We want some randomizing so keep tracks for later usage
            spotify_id = choice(spotify_ids)
            track_ids += spotify_ids
            payload = self.get_track_payload(spotify_id, cancellable)
            lollypop_payload = self.lollypop_album_payload(payload["album"])
            item = self.save_album_payload_to_db(lollypop_payload,
                                                 storage_type,
                                                 True,
                                                 cancellable)
            lollypop_payload = self.lollypop_track_payload(payload)
            self.save_track_payload_to_db(lollypop_payload,
                                          item,
                                          storage_type,
                                          True,
                                          cancellable)
        shuffle(track_ids)
        for spotify_id in track_ids:
            payload = self.get_track_payload(spotify_id, cancellable)
            lollypop_payload = self.lollypop_album_payload(payload["album"])
            item = self.save_album_payload_to_db(lollypop_payload,
                                                 storage_type,
                                                 True,
                                                 cancellable)
            lollypop_payload = self.lollypop_track_payload(payload)
            self.save_track_payload_to_db(lollypop_payload,
                                          item,
                                          storage_type,
                                          True,
                                          cancellable)
        emit_signal(self, "finished")

    def get_similar_artist_ids(self, artist_names, cancellable):
        """
            Get similar artists
            @param artist_names as [str]
            @param cancellable as Gio.Cancellable
            @return [str] as [spotify ids]
        """
        result = []
        for artist_name in artist_names:
            if cancellable.is_cancelled():
                return []
            spotify_id = self.get_artist_id(artist_name, cancellable)
            if spotify_id is None:
                continue
            if cancellable.is_cancelled():
                return []
            result += self.__get_similar_artists_from_spotify_id(spotify_id,
                                                                 cancellable)
        if result:
            Logger.info("Found similar artists with SpotifySimilars")
        return [spotify_id for (spotify_id, name, uri) in result]

    def get_similar_artists(self, artist_names, cancellable):
        """
            Get similar artists
            @param artist_names as [str]
            @param cancellable as Gio.Cancellable
            @return [(str, str)] as [(artist_name, cover_uri)]
        """
        result = []
        for artist_name in artist_names:
            if cancellable.is_cancelled():
                return []
            spotify_id = self.get_artist_id(artist_name, cancellable)
            if spotify_id is None:
                continue
            if cancellable.is_cancelled():
                return []
            result += self.__get_similar_artists_from_spotify_id(spotify_id,
                                                                 cancellable)
        result = [(name, uri) for (spotify_id, name, uri) in result]
        if result:
            Logger.info("Found similar artists with SpotifySimilars")
        return result

#######################
# PRIVATE             #
#######################
    def __get_similar_artists_from_spotify_id(self, spotify_id, cancellable):
        """
           Get similar artists from spotify id
           @param spotify_id as str
           @param cancellable as Gio.Cancellable
           @return [(str, str, str)] : list of (spotify_id, artist, cover_uri)
        """
        artists = []
        try:
            token = App().ws_director.token_ws.get_token("SPOTIFY",
                                                         cancellable)
            bearer = "Bearer %s" % token
            headers = [("Authorization", bearer)]
            uri = "https://api.spotify.com/v1/artists/%s/related-artists" %\
                spotify_id
            (status,
             data) = App().task_helper.load_uri_content_sync_with_headers(
                    uri, headers, cancellable)
            if status:
                decode = json.loads(data.decode("utf-8"))
                for item in decode["artists"]:
                    try:
                        image_uri = item["images"][1]["url"]
                    except:
                        image_uri = None
                    artists.append((item["id"],
                                    item["name"],
                                    image_uri))
        except:
            Logger.error(
                "SpotifySimilars::__get_similar_artists_from_spotify_id(): %s",
                spotify_id)
        return artists



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

import itertools

from lollypop.sqlcursor import SqlCursor
from lollypop.define import App, Type, OrderBy, LovedFlags
from lollypop.utils import get_network_available, sql_escape


class GenresDatabase:
    """
        Genres database helper
    """

    def __init__(self, db):
        """
            Init genres database object
            @param db as Database
        """
        self.__db = db

    def add(self, name):
        """
            Add a new genre to database
            @param name as string
            @return inserted rowid as int
            @warning: commit needed
        """
        with SqlCursor(self.__db, True) as sql:
            result = sql.execute("INSERT INTO genres (name) VALUES (?)",
                                 (name,))
            return result.lastrowid

    def get_id(self, name):
        """
            Get genre id for name
            @param name as string
            @return genre id as int
        """
        with SqlCursor(self.__db) as sql:
            # Escape string to fix mixed tags:
            # Alternative Rock, Aternative-Rock, alternative rock
            result = sql.execute("SELECT rowid FROM genres\
                                  WHERE sql_escape(name)=?",
                                 (sql_escape(name),))
            v = result.fetchone()
            if v is not None:
                return v[0]
            return None

    def get_name(self, genre_id):
        """
            Get genre name for genre id
            @param genre_id as int
            @return str
        """
        with SqlCursor(self.__db) as sql:
            result = sql.execute("SELECT name FROM genres\
                                  WHERE rowid=?", (genre_id,))
            v = result.fetchone()
            if v is not None:
                return v[0]
            return None

    def get_album_ids(self, ignore=False):
        """
            Get all availables albums for genres
            @param genre_ids as [int]
            @param ignore as bool
            @return [int]
        """
        orderby = App().settings.get_enum("orderby")
        order = " ORDER BY genres.name, "
        if orderby == OrderBy.ARTIST_YEAR:
            order += " artists.sortname\
                     COLLATE NOCASE COLLATE LOCALIZED,\
                     albums.timestamp,\
                     albums.name\
                     COLLATE NOCASE COLLATE LOCALIZED"
        elif orderby == OrderBy.ARTIST_TITLE:
            order += " artists.sortname\
                     COLLATE NOCASE COLLATE LOCALIZED,\
                     albums.name\
                     COLLATE NOCASE COLLATE LOCALIZED"
        elif orderby == OrderBy.NAME:
            order += " albums.name\
                     COLLATE NOCASE COLLATE LOCALIZED"
        elif orderby == OrderBy.YEAR_DESC:
            order += " albums.timestamp DESC,\
                     albums.name\
                     COLLATE NOCASE COLLATE LOCALIZED"
        else:
            order += " albums.popularity DESC,\
                     albums.name\
                     COLLATE NOCASE COLLATE LOCALIZED"
        with SqlCursor(self.__db) as sql:
            filters = ()
            request = "SELECT albums.rowid\
                       FROM albums, album_genres, genres,\
                            album_artists, artists\
                       WHERE album_genres.album_id=albums.rowid\
                       AND album_genres.genre_id = genres.rowid\
                       AND album_artists.artist_id = artists.rowid\
                       AND album_artists.album_id = albums.rowid"
            if not get_network_available():
                request += " AND albums.synced!=%s" % Type.NONE
            if ignore:
                request += " AND not albums.loved & ?"
                filters += (LovedFlags.SKIPPED,)
            request += order
            result = sql.execute(request, filters)
            return list(itertools.chain(*result))

    def get(self):
        """
            Get all availables genres
            @return [(int, str, str)]
        """
        with SqlCursor(self.__db) as sql:
            result = sql.execute("SELECT DISTINCT\
                                  genres.rowid, genres.name, genres.name\
                                  FROM genres\
                                  WHERE EXISTS (\
                                    SELECT *\
                                    FROM album_genres, album_artists\
                                    WHERE album_genres.album_id=\
                                        album_artists.album_id AND\
                                        album_artists.artist_id != ? AND\
                                        album_genres.genre_id=genres.rowid)\
                                  ORDER BY genres.name\
                                  COLLATE NOCASE COLLATE LOCALIZED",
                                 (Type.COMPILATIONS,))
            return list(result)

    def get_ids(self):
        """
            Get all availables genres ids
            @return [id as int]
        """
        with SqlCursor(self.__db) as sql:
            result = sql.execute("SELECT DISTINCT genres.rowid\
                                  FROM genres\
                                  WHERE EXISTS (\
                                    SELECT *\
                                    FROM album_genres, album_artists\
                                    WHERE album_genres.album_id=\
                                        album_artists.album_id AND\
                                        album_artists.artist_id != ? AND\
                                        album_genres.genre_id=genres.rowid)\
                                  ORDER BY genres.name\
                                  COLLATE NOCASE COLLATE LOCALIZED",
                                 (Type.COMPILATIONS,))
            return list(itertools.chain(*result))

    def get_random(self):
        """
            Return a random genre
            @return [int]
        """
        with SqlCursor(self.__db) as sql:
            result = sql.execute("SELECT genres.rowid, genres.name\
                                  FROM genres\
                                  WHERE EXISTS (\
                                    SELECT albums.rowid\
                                    FROM albums, album_genres\
                                    WHERE albums.loved != -1 AND\
                                          albums.rowid =\
                                            album_genres.album_id AND\
                                          album_genres.genre_id =\
                                            genres.rowid)\
                                  ORDER BY random() LIMIT 1")
            genres = list(result)
            return genres[0] if genres else (None, "")

    def clean(self, commit=True):
        """
            Clean genres
            @param commit as bool
        """
        with SqlCursor(self.__db, commit) as sql:
            sql.execute("DELETE FROM genres WHERE genres.rowid NOT IN (\
                            SELECT album_genres.genre_id FROM album_genres)")
            sql.execute("DELETE FROM genres WHERE genres.rowid NOT IN (\
                            SELECT track_genres.genre_id FROM track_genres)")



# Copyright (c) 2014-2016 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# Copyright (C) 2011 kedals0@gmail.com
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib, Gst

from datetime import datetime
import socketserver
import threading
import os

from lollypop.define import Lp, Type
from lollypop.objects import Track
from lollypop.database_mpd import MpdDatabase
from lollypop.utils import get_ip


class MpdHandler(socketserver.StreamRequestHandler):
    # Delayed signals
    _PLCHANGES = ["add", "delete", "clear", "deleteid", "move",
                  "moveid", "load", "playlistadd"]

    def handle(self):
        """
            One function to handle them all
        """
        self.request.send("OK MPD 0.19.0\n".encode('utf-8'))
        self._idle_thread = None
        while self.server.running:
            msg = ""
            try:
                cmdlist = None
                cmds = []
                # check for delayed plchanges
                delayed = False
                # Read commands
                while True:
                    print("True")
                    data = self.rfile.readline().strip().decode("utf-8")
                    print("False")
                    if len(data) == 0:
                        raise IOError
                    if data == "command_list_ok_begin":
                        cmdlist = "list_ok"
                    elif data == "command_list_begin":
                        cmdlist = "list"
                    elif data == "command_list_end":
                        break
                    else:
                        cmds.append(data)
                        if not cmdlist:
                            break
                if not cmds:
                    continue
                try:
                    print(cmds, self)
                    for cmd in cmds:
                        command = cmd.split(' ')[0]
                        size = len(command) + 1
                        args = cmd[size:]
                        if command in self._PLCHANGES:
                            delayed = True
                        call = getattr(self, '_%s' % command)
                        result = call(args)
                        if result is None:
                            msg = None
                            break
                        else:
                            msg += result
                            if cmdlist == "list_ok":
                                msg += "list_OK\n"
                    if msg is None:
                        continue
                except Exception as e:
                    print("MpdHandler::handle(): ", cmds, e)
                msg += "OK\n"
                self.wfile.write(msg.encode("utf-8"))
                print(msg.encode("utf-8"), self)
                if delayed:
                    GLib.idle_add(Lp().playlists.emit,
                                  'playlist-changed',
                                  Type.MPD)
            except Exception as e:
                print(e)
                break

    def _add(self, cmd_args):
        """
            Add track to mpd playlist
            @syntax add filepath
            @param args as str
            @return msg as str
        """
        tracks = []
        arg = self._get_args(cmd_args)[0]
        try:
            arg = GLib.filename_from_uri(arg)[0]
        except:
            pass
        track_id = Lp().tracks.get_id_by_path(arg)
        if track_id is None:
            path = ""
            for musicpath in Lp().settings.get_music_paths():
                search = musicpath.replace("/", "_")
                if search in arg:
                    path = musicpath + arg.replace(search, path)
            if os.path.isdir(path):
                tracks_ids = Lp().tracks.get_ids_by_path(path)
                for track_id in tracks_ids:
                    tracks.append(Track(track_id))
            elif os.path.isfile(path):
                track_id = Lp().tracks.get_id_by_path(path)
                tracks.append(Track(track_id))
        else:
            tracks.append(Track(track_id))
        Lp().playlists.add_tracks(Type.MPD, tracks, False)
        return ""

    def _addid(self, cmd_args):
        """
            Add track to mpd playlist at pos
            @syntax add filepath
            @param args as str
            @return msg as str
        """
        arg = self._get_args(cmd_args)[0]
        self._add(cmd_args)
        try:
            arg = GLib.filename_from_uri(arg)[0]
        except:
            pass
        return "Id: %s" % Lp().tracks.get_id_by_path(arg)

    def _clear(self, cmd_args):
        """
            Clear mpd playlist
            @syntax clear
            @param args as str
            @return msg as str
        """
        Lp().playlists.clear(Type.MPD, False)
        Lp().player.populate_user_playlist_by_id(Type.NONE)
        GLib.idle_add(Lp().player.stop)
        Lp().player.current_track = Track()
        GLib.idle_add(Lp().player.emit, 'current-changed')
        return ""

    def _channels(self, cmd_args):
        return ""

    def _commands(self, cmd_args):
        """
            Send available commands
            @syntax commands
            @param args as str
            @return msg as str
        """
        msg = "command: add\ncommand: addid\ncommand: clear\
\ncommand: channels\ncommand: count\ncommand: currentsong\
\ncommand: delete\ncommand: deleteid\ncommand: idle\ncommand: noidle\
\ncommand: list\ncommand: listallinfo\ncommand: listplaylists\ncommand: lsinfo\
\ncommand: next\ncommand: outputs\ncommand: pause\ncommand: play\
\ncommand: playid\ncommand: playlistinfo\ncommand: plchanges\
\ncommand: plchangesposid\ncommand: prev\ncommand:find\ncommand:findadd\
\ncommand: replay_gain_status\
\ncommand: repeat\ncommand: seek\ncommand: seekid\ncommand: search\
\ncommand: setvol\ncommand: stats\ncommand: status\ncommand: sticker\
\ncommand: stop\ncommand: tagtypes\ncommand: update\ncommand: urlhandlers\n"
        return msg

    def _count(self, cmd_args):
        """
            Send lollypop current song
            @syntax count tag
            @param args as str
            @return msg as str
        """
        args = self._get_args(cmd_args)
        # Search for filters
        i = 0
        artist = artist_id = year = album = genre = genre_id = None
        while i < len(args) - 1:
            if args[i].lower() == 'album':
                album = args[i+1]
            elif args[i].lower() == 'artist':
                artist = args[i+1]
            elif args[i].lower() == 'genre':
                genre = args[i+1]
            elif args[i].lower() == 'date':
                date = args[i+1]
            i += 2

        # Artist have albums with different dates so
        # we do not want None in year
        if artist_id is not None or album is not None:
            try:
                year = int(date)
            except:
                year = None
        else:
            year = Type.NONE

        if genre is not None:
            genre_id = Lp().genres.get_id(genre)
        if artist is not None:
            artist_id = Lp().artists.get_id(artist)

        (songs, playtime) = self.server.mpddb.count(album, artist_id,
                                                    genre_id, year)
        msg = "songs: %s\nplaytime: %s\n" % (songs, playtime)
        return msg

    def _currentsong(self, cmd_args):
        """
            Send lollypop current song
            @syntax currentsong
            @param args as str

            @return msg as str
        """
        if Lp().player.current_track.id is not None:
            msg = self._string_for_track_id(Lp().player.current_track.id)
        else:
            msg = ""
        return msg

    def _delete(self, cmd_args):
        """
            Delete track from playlist
            @syntax delete position
            @param args as str
            @return msg as str
        """
        arg = self._get_args(cmd_args)[0]
        # Check for a range
        try:
            splited = arg.split(':')
            start = int(splited[0]) - 1
            end = int(splited[1])
        except:
            start = int(arg) - 1
            end = int(arg)
        tracks_ids = Lp().playlists.get_tracks_ids(Type.MPD)
        tracks = []
        for i in range(start, end):
            track_id = tracks_ids[i]
            tracks.append(Track(track_id))
        Lp().playlists.remove_tracks(Type.MPD, tracks, False)
        return ""

    def _deleteid(self, cmd_args):
        """
            Delete track from playlist
            @syntax delete track_id
            @param args as str

            @return msg as str
        """
        arg = self._get_args(cmd_args)
        Lp().playlists.remove_tracks(Type.MPD, [Track(int(arg[0]))], False)
        return ""

    def _find(self, cmd_args):
        """
            find tracks
            @syntax find filter value
            @param args as str

            @return msg as str
        """
        msg = ""
        idx = 0
        for track_id in self._find_tracks(cmd_args):
            msg += self._string_for_track_id(track_id, idx)
            idx += 1
        return msg

    def _findadd(self, cmd_args):
        """
            Find tracks and add them to playlist
            @syntax findadd filter value
            @param args as str
            @return msg as str
        """
        tracks = []
        for track_id in self._find_tracks(cmd_args):
            tracks.append(Track(track_id))
        if tracks:
            Lp().playlists.add_tracks(Type.MPD, tracks, False)
        return ""

    def _idle(self, cmd_args):
        """
            Idle waiting for changes
            @syntax idle type (type not implemented here)
            @param args as str
            @return msg as str
        """
        if self._idle_thread is None:
            self._idle_thread = threading.Thread(target=self._idle_threaded,
                                                 args=(cmd_args,))
            self._idle_thread.daemon = True
            self._idle_thread.start()
        return None

    def _idle_threaded(self, cmd_args):
        """
            Same as previous, thread safe
        """
        msg = ''
        args = self._get_args(cmd_args)
        if args:
            self.server.idle_wanted_strings = []
            for arg in args:
                self.server.idle_wanted_strings.append(arg)
        else:
            self.server.idle_wanted_strings = ["stored_playlist",
                                               "player", "playlist", "options"]
        self.server.event.clear()
        # We handle notifications directly if something in queue
        if not self.server.idle_strings:
            self.server.event.wait()
        # Handle new notifications
        for string in self.server.idle_strings:
            msg += "changed: %s\n" % string
        if msg != "":
            msg += "OK\n"
            try:
                self.request.send(msg.encode("utf-8"))
            except:
                pass
            print("IDLE", msg.encode("utf-8"), self)
        self.server.idle_wanted_strings = []
        self.server.idle_strings = []
        self._idle_thread = None

    def _noidle(self, cmd_args):
        """
            Stop idle
            @syntax noidle
            @param args as str
            @return msg as str
        """
        msg = ""
        # Handle new notifications
        for string in self.server.idle_strings:
            msg += "changed: %s\n" % string
        msg += "OK\n"
        self.request.send(msg.encode("utf-8"))
        print("NOIDLE", msg.encode("utf-8"), self)
        self.server.idle_wanted_strings = []
        self.server.idle_strings = []
        self._idle_thread = None
        self.server.event.set()
        return None

    def _list(self, cmd_args):
        """
            List objects
            @syntax list what [filter value...] or list album artist_name
            @param args as str
            @return msg as str
        """
        msg = ""
        args = self._get_args(cmd_args)
        # Search for filters
        if len(args) == 2:
            i = 0
        else:
            i = 1
        artist = artist_id = None
        album = None
        genre = genre_id = None
        date = ''
        while i < len(args) - 1:
            if args[i].lower() == 'album':
                if i % 2:
                    album = args[i+1]
                else:
                    artist = args[i+1]
            elif args[i].lower() == 'artist' or\
                    args[i].lower() == 'albumartist':
                artist = args[i+1]
            elif args[i].lower() == 'genre':
                genre = args[i+1]
            elif args[i].lower() == 'date':
                date = args[i+1]
            i += 2

        try:
            year = int(date)
        except:
            year = Type.NONE

        if genre is not None:
            genre_id = Lp().genres.get_id(genre)
        if artist is not None:
            artist_id = Lp().artists.get_id(artist)

        if args[0].lower() == 'file':
            for path in self.server.mpddb.get_tracks_paths(album, artist_id,
                                                           genre_id, year):
                msg += "File: "+path+"\n"
        if args[0].lower() == 'album':
            print(artist_id)
            for album in self.server.mpddb.get_albums_names(artist_id,
                                                            genre_id, year):
                msg += "Album: "+album+"\n"
        elif args[0].lower() == 'artist':
            for artist in self.server.mpddb.get_artists_names(genre_id):
                msg += "Artist: "+artist+"\n"
        elif args[0].lower() == 'genre':
            results = Lp().genres.get_names()
            for name in results:
                msg += "Genre: "+name+"\n"
        elif args[0].lower() == 'date':
            for year in self.server.mpddb.get_albums_years(album, artist_id,
                                                           genre_id):
                msg += "Date: "+str(year)+"\n"
        return msg

    def _listall(self, cmd_args):
        """
            List all tracks
            @syntax listall
            @param args as str
            @return msg as str
        """
        return ""

    def _listallinfo(self, cmd_args):
        """
            List all tracks
            @syntax listallinfo
            @param args as str
            @return msg as str
        """
        i = 0
        msg = ""
        for (path, artist, album, album_artist,
             title, date, genre, time,
             track_id, pos) in self.server.mpddb.listallinfos():
            msg += "file: %s\nArtist: %s\nAlbum: %s\nAlbumArtist: %s\
\nTitle: %s\nDate: %s\nGenre: %s\nTime: %s\nId: %s\nPos: %s\nTrack: %s\n" % (
                                        path,
                                        artist,
                                        album,
                                        album_artist,
                                        title,
                                        date,
                                        genre,
                                        time,
                                        track_id,
                                        pos,
                                        pos)
            if i > 1000:
                self.request.send(msg.encode("utf-8"))
                msg = ""
                i = 0
            else:
                i += 1
        return msg

    def _listplaylistinfo(self, cmd_args):
        """
            List playlist information
            @syntax listplaylistinfo name
            @param args as str
            @return msg as str
        """
        arg = self._get_args(cmd_args)[0]
        playlist_id = Lp().playlists.get_id(arg)
        msg = ""
        idx = 0
        for track_id in Lp().playlists.get_tracks_ids(playlist_id):
            msg += self._string_for_track_id(track_id, idx)
            idx += 1
        return msg

    def _listplaylists(self, cmd_args):
        """
            Send available playlists
            @syntax listplaylists
            @param args as str
            @return msg as str
        """
        dt = datetime.utcnow()
        dt = dt.replace(microsecond=0)
        msg = ""
        for (playlist_id, name) in Lp().playlists.get():
            msg += "playlist: %s\nLast-Modified: %s\n" % (
                                                      name,
                                                      '%sZ' % dt.isoformat())
        return msg

    def _load(self, cmd_args):
        """
            Load playlist
            @syntax load name
            @param args as str
            @return msg as str
        """
        arg = self._get_args(cmd_args)[0]
        playlist_id = Lp().playlists.get_id(arg)
        tracks = []
        tracks_ids = Lp().playlists.get_tracks_ids(playlist_id)
        for track_id in tracks_ids:
            tracks.append(Track(track_id))
        Lp().playlists.add_tracks(Type.MPD, tracks, False)
        self.server.init_player_playlist()
        GLib.idle_add(Lp().player.load_in_playlist, tracks_ids[0])
        return ""

    def _lsinfo(self, cmd_args):
        """
            List directories and files
            @syntax lsinfo path
            @param args as str
            @return msg as str
        """
        msg = ""
        args = self._get_args(cmd_args)
        if not args:
            arg = ""
        else:
            arg = args[0]
            if arg == "/":
                arg = ""
        results = []
        root = ""
        if arg == "":
            for path in Lp().settings.get_music_paths():
                results.append((True, path.replace("/", "_")))
        else:
            splited = arg.split("/")
            root = None
            for path in Lp().settings.get_music_paths():
                if path.replace("/", "_") in [arg, splited[0]]:
                    root = path + "/" + "/".join(splited[1:])
                    break
            if root is not None:
                for entry in os.listdir(root):
                    if os.path.isdir(root+"/"+entry):
                        results.append((True, entry))
                    else:
                        results.append((False, entry))
        i = 0
        for (d, path) in results:
            relative = path.replace(root, '', 1)
            if d:
                if arg == "":
                    msg += "directory: %s\n" % relative
                else:
                    msg += "directory: %s/%s\n" % (arg, relative)
            elif Lp().tracks.get_id_by_path(root+"/"+relative) is not None:
                msg += "file: %s/%s\n" % (arg, relative)
            if i > 100:
                self.request.send(msg.encode("utf-8"))
                msg = ""
                i = 0
            i += 1
        return msg

    def _next(self, cmd_args):
        """
            Send output
            @syntax next
            @param args as str
            @return msg as str
        """
        # Make sure we have a playlist loaded in player
        if not Lp().player.is_party():
            if Lp().player.get_user_playlist_id() != Type.MPD or not\
               Lp().player.get_user_playlist():
                self.server.init_player_playlist()
            tracks_ids = Lp().playlists.get_tracks_ids(Type.MPD)
            if tracks_ids and Lp().player.current_track.id not in tracks_ids:
                GLib.idle_add(Lp().player.load_in_playlist, tracks_ids[0])
                return ""
        GLib.idle_add(Lp().player.next)
        return ""

    def _move(self, cmd_args):
        """
            Move range in playlist
            @syntax move position destination
            @param args as str
            @return msg as str
        """
        # TODO implement range
        tracks_ids = Lp().playlists.get_tracks_ids(Type.MPD)
        arg = self._get_args(cmd_args)
        orig = int(arg[0])
        dst = int(arg[1])
        if orig != dst:
            track_id = tracks_ids[orig]
            del tracks_ids[orig]
            tracks_ids.insert(dst, track_id)
            Lp().playlists.clear(Type.MPD, False)
            tracks = []
            for track_id in tracks_ids:
                tracks.append(Track(track_id))
            Lp().player.populate_user_playlist_by_id(Type.NONE)
            Lp().playlists.add_tracks(Type.MPD, tracks, False)
        return ""

    def _moveid(self, cmd_args):
        """
            Move id in playlist
            @syntax move track_id destination
            @param args as str
            @return msg as str
        """
        try:
            tracks_ids = Lp().playlists.get_tracks_ids(Type.MPD)
            arg = self._get_args(cmd_args)
            track_id = int(arg[0])
            orig = tracks_ids.index(track_id)
            dst = int(arg[1])
            del tracks_ids[orig]
            tracks_ids.insert(dst, track_id)

            Lp().playlists.clear(Type.MPD)
            tracks = []
            for track_id in tracks_ids:
                tracks.append(Track(track_id))
            Lp().player.populate_user_playlist_by_id(Type.NONE)
            Lp().playlists.add_tracks(Type.MPD, tracks, False)
        except:
            pass
        return ""

    def _outputs(self, cmd_args):
        """
            Send output
            @syntax outputs
            @param args as str
            @return msg as str
        """
        msg = "outputid: 0\noutputname: null\noutputenabled: 1\n"
        return msg

    def _pause(self, cmd_args):
        """
            Pause track
            @syntax pause [1|0]
            @param args as str
            @return msg as str
        """
        print("debut")
        try:
            args = self._get_args(cmd_args)
            if args[0] == "0":
                GLib.idle_add(Lp().player.play)
            else:
                GLib.idle_add(Lp().player.pause)
        except:
            GLib.idle_add(Lp().player.play_pause)
        print('fin')
        return ""

    def _play(self, cmd_args):
        """
            Play track
            @syntax play [position|-1]
            @param args as str
            @return msg as str
        """
        if Lp().player.is_party():
            # Force player to not load albums
            Lp().player.current_track.id = None
            GLib.idle_add(Lp().player.set_party, False)
        self.server.init_player_playlist()
        try:
            arg = int(self._get_args(cmd_args)[0])
            currents = Lp().player.get_user_playlist()
            if currents:
                track = currents[arg]
                GLib.idle_add(Lp().player.load_in_playlist, track.id)
        except:
            arg = -1
        if Lp().player.get_status() == Gst.State.PAUSED:
            GLib.idle_add(Lp().player.play)
        elif Lp().player.get_status() == Gst.State.NULL:
            if Lp().player.current_track.id is not None:
                GLib.idle_add(Lp().player.play)
            else:
                currents = Lp().player.get_user_playlist()
                if currents:
                    track = currents[0]
                    GLib.idle_add(Lp().player.load_in_playlist, track.id)
        return ""

    def _playid(self, cmd_args):
        """
            Play track
            @syntax play [track_id|-1]
            @param args as str
            @return msg as str
        """
        if Lp().player.is_party():
            # Force player to not load albums
            Lp().player.current_track.id = None
            GLib.idle_add(Lp().player.set_party, False)
        self.server.init_player_playlist()
        try:
            arg = int(self._get_args(cmd_args)[0])
            GLib.idle_add(Lp().player.load_in_playlist, arg)
        except:
            if Lp().player.get_status() == Gst.State.PAUSED:
                GLib.idle_add(Lp().player.play)
            elif Lp().player.get_status() == Gst.State.NULL:
                if Lp().player.current_track.id is not None:
                    GLib.idle_add(Lp().player.play)
                else:
                    currents = Lp().player.get_user_playlist()
                    if currents:
                        track = currents[0]
                        GLib.idle_add(Lp().player.load_in_playlist, track.id)
        return ""

    def _playlistadd(self, cmd_args):
        """
            Add a new playlist
            @syntax playlistadd name
            @param args as str
            @return msg as str
        """
        args = self._get_args(cmd_args)
        playlist_id = Lp().playlists.get_id(args[0])
        tracks = []
        if not Lp().playlists.exists(playlist_id):
            Lp().playlists.add(args[0])
            playlist_id = Lp().playlists.get_id(args[0])
        for arg in args[1:]:
            track_id = Lp().tracks.get_id_by_path(arg)
            tracks.append(Track(track_id))
        if tracks:
            Lp().playlists.add_tracks(playlist_id, tracks, False)
        return ""

    def _playlistid(self, cmd_args):
        """
            Send information about current playlist
            @param playlistid
            @param args as str
            @return msg as str
        """
        msg = ""
        try:
            track_id = int(self._get_args(cmd_args))
            msg += self._string_for_track_id(track_id)
        except:
            currents = Lp().playlists.get_tracks_ids(Type.MPD)
            if Lp().player.is_party():
                currents.insert(0, Lp().player.current_track.id)
                if Lp().player.prev_track.id is not None:
                    currents.insert(0, Lp().player.prev_track.id)
                if Lp().player.next_track.id is not None:
                    currents.append(Lp().player.next_track.id)
            for track_id in currents:
                msg += self._string_for_track_id(track_id)
        return msg

    def _playlistinfo(self, cmd_args):
        """
            Send information about current playlist
            @syntax playlistinfo [[pos]|[start:end]]
            @param playlistinfo
            @param args as str
            @return msg as str
        """
        msg = ""
        try:
            arg = self._get_args(cmd_args)[0]
        except:
            arg = None
        start = end = pos = None
        if arg is not None:
            # Check for a range
            try:
                splited = arg.split(':')
                start = int(splited[0])
                end = int(splited[1])
            except:
                start = end = None
                pos = int(arg)
        currents = Lp().playlists.get_tracks_ids(Type.MPD)
        if Lp().player.is_party():
            currents.insert(0, Lp().player.current_track.id)
            if Lp().player.prev_track.id is not None:
                currents.insert(0, Lp().player.prev_track.id)
            if Lp().player.next_track.id is not None:
                currents.append(Lp().player.next_track.id)
        i = 0
        for track_id in currents:
            if (start is not None and start <= i <= end) or\
               (pos is not None and pos == i) or\
               (start == end == pos is None):
                msg += self._string_for_track_id(track_id)
            i += 1
        return msg

    def _plchanges(self, cmd_args):
        """
            Displays changed songs currently in the playlist since version
            @syntax plchanges version
            @param args as str
            @return msg as str
        """
        msg = ""
        version = int(self._get_args(cmd_args)[0])
        i = 0
        currents = list(Lp().playlists.get_tracks_ids(Type.MPD))
        if Lp().player.is_party():
            currents.insert(0, Lp().player.current_track.id)
            if Lp().player.prev_track.id is not None:
                currents.insert(0, Lp().player.prev_track.id)
            if Lp().player.next_track.id is not None:
                currents.append(Lp().player.next_track.id)
        previous = list(self.server.playlist[version])
        while currents:
            current = currents.pop(0)
            try:
                prev = previous.pop(0)
            except:
                prev = Type.NONE
            if current != prev:
                msg += self._string_for_track_id(current)
                if i > 100:
                    self.request.send(msg.encode("utf-8"))
                    msg = ""
                    i = 0
                else:
                    i += 1
        return msg

    def _plchangesposid(self, cmd_args):
        """
            Displays changed songs currently in the playlist since version
            @param plchangesposid version
            @param args as str
            @return msg as str
        """
        i = 0  # Sending index
        idx = 0  # Track index
        msg = ""
        version = int(self._get_args(cmd_args)[0])
        currents = list(Lp().playlists.get_tracks_ids(Type.MPD))
        if Lp().player.is_party():
            currents.insert(0, Lp().player.current_track.id)
            if Lp().player.prev_track.id is not None:
                currents.insert(0, Lp().player.prev_track.id)
            if Lp().player.next_track.id is not None:
                currents.append(Lp().player.next_track.id)
        previous = list(self.server.playlist[version])
        while currents:
            current = currents.pop(0)
            try:
                prev = previous.pop(0)
            except:
                prev = Type.NONE
            if current != prev:
                msg += "cpos: %s\nId: %s\n" % (idx, current)
                if i > 100:
                    self.request.send(msg.encode("utf-8"))
                    msg = ""
                    i = 0
                else:
                    i += 1
            idx += 1
        return msg

    def _previous(self, cmd_args):
        """
            Send output
            @syntax previous
            @param args as str
            @return msg as str
        """
        # Make sure we have a playlist loaded in player
        if not Lp().player.is_party():
            if Lp().player.get_user_playlist_id() != Type.MPD or not\
               Lp().player.get_user_playlist():
                self.server.init_player_playlist()
            tracks_ids = Lp().playlists.get_tracks_ids(Type.MPD)
            if tracks_ids and Lp().player.current_track.id not in tracks_ids:
                GLib.idle_add(Lp().player.load_in_playlist, tracks_ids[0])
                return ""
        GLib.idle_add(Lp().player.prev)
        return ""

    def _random(self, cmd_args):
        """
            Set player random, as MPD can't handle all lollypop random modes,
            set party mode
            @syntax random [1|0]
            @param args as str
            @return msg as str
        """
        args = self._get_args(cmd_args)
        GLib.idle_add(Lp().player.set_party, bool(int(args[0])))
        return ""

    def _replay_gain_status(self, cmd_args):
        """
            Send output
            @syntax replay_gain_status
            @param args as str
            @return msg as str
        """
        msg = "replay_gain_mode: on\n"
        return msg

    def _repeat(self, cmd_args):
        """
            Ignore
            @param args as str
            @return msg as str
        """
        return ""

    def _seek(self, cmd_args):
        """
           Seek current
           @syntax seek position
           @param args as str
           @return msg as str
        """
        args = self._get_args(cmd_args)
        seek = int(args[1])
        GLib.idle_add(Lp().player.seek, seek)
        return ""

    def _seekid(self, cmd_args):
        """
            Seek track id
            @syntax seekid track_id position
            @param args as str
            @return msg as str
        """
        args = self._get_args(cmd_args)
        track_id = int(args[0])
        seek = int(args[1])
        if track_id == Lp().player.current_track.id:
            GLib.idle_add(Lp().player.seek, seek)
        return ""

    def _search(self, cmd_args):
        """
            Send stats about db
            @syntax search what value
            @param args as str
            @return msg as str
        """
        msg = ""
        args = self._get_args(cmd_args)
        # Search for filters
        i = 0
        artist = artist_id = None
        album = None
        genre = genre_id = None
        date = ''
        while i < len(args) - 1:
            if args[i].lower() == 'album':
                album = args[i+1]
            elif args[i].lower() == 'artist' or\
                    args[i].lower() == 'albumartist':
                artist = args[i+1]
            elif args[i].lower() == 'genre':
                genre = args[i+1]
            elif args[i].lower() == 'date':
                date = args[i+1]
            i += 2

        try:
            year = int(date)
        except:
            year = Type.NONE

        if genre is not None:
            genre_id = Lp().genres.get_id(genre)
        if artist is not None:
            artist_id = Lp().artists.get_id(artist)

        for track_id in self.server.mpddb.get_tracks_ids(album, artist_id,
                                                         genre_id, year):
            msg += self._string_for_track_id(track_id)
        return msg

    def _setvol(self, cmd_args):
        """
            Send stats about db
            @syntax setvol value
            @param args as str
            @return msg as str
        """
        args = self._get_args(cmd_args)
        vol = float(args[0])
        Lp().player.set_volume(vol/100)
        return ""

    def _stats(self, cmd_args):
        """
            Send stats about db
            @syntax stats
            @param args as str
            @return msg as str
        """
        artists = Lp().artists.count()
        albums = Lp().albums.count()
        tracks = Lp().tracks.count()
        msg = "artists: %s\nalbums: %s\nsongs: %s\nuptime: 0\
\nplaytime: 0\ndb_playtime: 0\ndb_update: %s\n" % \
            (artists, albums, tracks,
             Lp().settings.get_value('db-mtime').get_int32())
        return msg

    def _status(self, cmd_args):
        """
            Send lollypop status
            @syntax status
            @param args as str
            @return msg as str
        """
        if Lp().player.is_party():
            pos = 1
            playlistlength = 3
        else:
            pos = Lp().playlists.get_position(Type.MPD,
                                              Lp().player.current_track.id)
            playlistlength = len(Lp().playlists.get_tracks(Type.MPD))
        msg = "volume: %s\nrepeat: %s\nrandom: %s\
\nsingle: %s\nconsume: %s\nplaylist: %s\
\nplaylistlength: %s\nstate: %s\
\nbitrate: 0\naudio: 44100:24:2\nmixrampdb: 0.000000\n" % (
                                   int(Lp().player.get_volume()*100),
                                   1,
                                   int(Lp().player.is_party()),
                                   1,
                                   1,
                                   self.server.playlist_version,
                                   playlistlength,
                                   self._get_status(),
                                   )
        if self._get_status() != 'stop':
            elapsed = Lp().player.get_position_in_track() / 1000000 / 60
            time = Lp().player.current_track.duration
            songid = Lp().player.current_track.id
            msg += "song: %s\nsongid: %s\ntime: %s:%s\nelapsed: %s\n" % (
                                       pos,
                                       songid,
                                       int(elapsed),
                                       time,
                                       int(elapsed))
        return msg

    def _sticker(self, cmd_args):
        """
            Send stickers
            @syntax sticker [get|set] song rating
            @param args as str
            @return msg as str
        """
        args = self._get_args(cmd_args)
        msg = ""
        print(args)
        if args[0] == "get" and args[1] == "song" and\
                args[3] == "rating":
            track_id = Lp().tracks.get_id_by_path(args[2])
            track = Track(track_id)
            msg = "sticker: rating=%s\n" % int((track.get_popularity()-0.5)*2)
        elif args[0] == "set" and args[1] == "song" and\
                args[3] == "rating":
            track_id = Lp().tracks.get_id_by_path(args[2])
            track = Track(track_id)
            track.set_popularity(int(args[4])/2)
        return msg

    def _stop(self, cmd_args):
        """
            Stop player
            @syntax stop
            @param args as str
            @return msg as str
        """
        GLib.idle_add(Lp().player.stop)
        return ""

    def _tagtypes(self, cmd_args):
        """
            Send available tags
            @syntax tagtypes
            @param args as str
            @return msg as str
        """
        msg = "tagtype: Artist\ntagtype: Album\ntagtype: Title\
\ntagtype: Track\ntagtype: Name\ntagtype: Genre\ntagtype: Date\
\ntagtype: Performer\ntagtype: Disc\n"
        return msg

    def _update(self, cmd_args):
        """
            Update database
            @syntax update
            @param args as str
            @return msg as str
        """
        Lp().window.update_db()
        return ""

    def _urlhandlers(self, cmd_args):
        """
            Send url handlers
            @syntax urlhandlers
            @param args as str
            @return msg as str
        """
        msg = "handler: http\n"
        return msg

    def _string_for_track_id(self, track_id, index=Type.NONE):
        """
            Get mpd protocol string for track id
            @param track_id as int
            @param track index as int
            @return str
        """
        if track_id is None:
            msg = ""
        else:
            track = Track(track_id)
            if index == Type.NONE:
                index = 1
                if Lp().player.is_party():
                    tracks_ids = [Lp().player.prev_track.id,
                                  Lp().player.current_track.id,
                                  Lp().player.next_track.id]
                    index = tracks_ids.index(track_id)
                else:
                    tracks_ids = Lp().playlists.get_tracks_ids(Type.MPD)
                    try:
                        index = tracks_ids.index(track_id) + 1
                    except:
                        pass
            msg = "file: %s\nArtist: %s\nAlbum: %s\nAlbumArtist: %s\
\nTitle: %s\nDate: %s\nGenre: %s\nTime: %s\nId: %s\nPos: %s\nTrack: %s\n" % (
                     track.path,
                     track.artist,
                     track.album.name,
                     track.album_artist,
                     track.name,
                     track.album.year,
                     track.genre,
                     track.duration,
                     track.id,
                     index,
                     index)
        return msg

    def _get_status(self):
        """
            Player status
            @return str
        """
        state = Lp().player.get_status()
        if state == Gst.State.PLAYING:
            return 'play'
        elif state == Gst.State.PAUSED:
            return 'pause'
        else:
            return 'stop'

    def _get_args(self, args):
        """
            Get args from string
            @param args as str
            @return args as [str]
        """
        ret = []
        for arg in args.split('"'):
            # Ignore bad args
            if arg.strip() == "":
                continue
            # Some client do not put "" for all args
            elif len(arg) > 0 and (arg[-1] == " " or
                                   arg[0] == " "):
                for subarg in arg.split(' '):
                    if subarg.strip() != "":
                        ret.append(subarg)
            else:
                ret.append(arg)
        return ret

    def _find_tracks(self, cmd_args):
        """
            find tracks
            @syntax find filter value
            @param args as str

        """
        tracks = []
        args = self._get_args(cmd_args)
        # Search for filters
        i = 0
        track_position = None
        artist = artist_id = None
        album = None
        genre = genre_id = None
        date = ''
        while i < len(args) - 1:
            if args[i].lower() == 'album':
                album = args[i+1]
            elif args[i].lower() == 'artist' or\
                    args[i].lower() == 'albumartist':
                artist = args[i+1]
            elif args[i].lower() == 'genre':
                genre = args[i+1]
            elif args[i].lower() == 'date':
                date = args[i+1]
            elif args[i].lower() == 'track':
                track_position = args[i+1]
            i += 2

        try:
            year = int(date)
        except:
            year = Type.NONE

        if genre is not None:
            genre_id = Lp().genres.get_id(genre)
        if artist is not None:
            artist_id = Lp().artists.get_id(artist)

        # We search for tracks and filter on position
        for track_id in self.server.mpddb.get_tracks_ids(album, artist_id,
                                                         genre_id, year):
            track_id_position = None
            if track_position is not None:
                track_id_position = Lp().tracks.get_position(track_id)
            if track_id_position == track_position:
                tracks.append(track_id)
        return tracks


class MpdServer(socketserver.ThreadingMixIn, socketserver.TCPServer):
    """
        Create a MPD server.
    """

    def __init__(self, eth, port=6600):
        """
            Init server
            @param eth as string
            @param port as int
        """
        self.event = None
        self.mpddb = MpdDatabase()
        self.playlist = {}
        self.playlist_version = 0
        self.idle_wanted_strings = []
        self.idle_strings = []
        try:
            # Set initial playlist version
            self.playlist[0] = []
            socketserver.TCPServer.allow_reuse_address = True
            # Get ip for interface
            ip = ""
            if eth != "":
                ip = get_ip(eth)
            socketserver.TCPServer.__init__(self, (ip, port), MpdHandler)
        except Exception as e:
            print("MpdServer::__init__(): %s" % e)

    def run(self, e):
        """
            Run MPD server in a blocking way.
            @param e as threading.Event
        """
        try:
            self._connect()
            self.event = e
            self.serve_forever()
            self._connect(False)
        except Exception as e:
            print("MpdServer::run(): %s" % e)

    def init_player_playlist(self):
        """
            Init player playlist if needed
        """
        if not Lp().player.is_party():
            Lp().player.populate_user_playlist_by_id(Type.MPD)

#######################
# PRIVATE             #
#######################
    def _connect(self, connect=True):
        """
            Connect or disconnect signals
        """
        if connect:
            self._signal1 = Lp().player.connect('current-changed',
                                                self._on_current_changed)
            self._signal2 = Lp().player.connect('status-changed',
                                                self._on_status_changed)
            self._signal3 = Lp().player.connect('seeked',
                                                self._on_position_changed)
            self._signal4 = Lp().player.connect('party-changed',
                                                self._on_party_changed)
            self._signal5 = Lp().playlists.connect('playlist-changed',
                                                   self._on_playlist_changed)
        else:
            Lp().player.disconnect(self._signal1)
            Lp().player.disconnect(self._signal2)
            Lp().player.disconnect(self._signal3)
            Lp().player.disconnect(self._signal4)
            Lp().playlists.disconnect(self._signal5)

    def _on_current_changed(self, player):
        """
            Add player to idle
            @param player as Player
        """
        changed = False
        if "player" in self.idle_wanted_strings:
            self.idle_strings.append("player")
            changed = True
        # We want to add party song to playlist
        if player.is_party() and "playlist" in self.idle_wanted_strings:
            self.idle_strings.append("playlist")
            self.playlist_version += 1
            self.playlist[self.playlist_version] = []
            if Lp().player.prev_track.id is not None:
                self.playlist[self.playlist_version] = [
                                                Lp().player.prev_track.id]
            self.playlist[self.playlist_version] = [
                                                Lp().player.current_track.id]
            if Lp().player.next_track.id is not None:
                self.playlist[self.playlist_version] = [
                                                Lp().player.next_track.id]
            changed = True
        if changed:
            self.event.set()

    def _on_status_changed(self, player, data=None):
        """
            Add player to idle
            @param player as Player
        """
        if "player" in self.idle_wanted_strings:
            self.idle_strings.append("player")
            self.event.set()

    def _on_position_changed(self, player, data=None):
        """
            Add player to idle
            @param player as Player
        """
        # Player may be in pause so wait for playback
        if player.get_status() == Gst.State.PAUSED:
            GLib.idle_add(self._on_position_changed, player, data)
        elif "player" in self.idle_wanted_strings:
            self.idle_strings.append("player")
            self.event.set()

    def _on_party_changed(self, player, enabled):
        """
            Add options to idle
            @param player as Player
            @param enabled as bool
        """
        Lp().playlists.clear(Type.MPD, False)
        if "options" in self.idle_wanted_strings:
            self.idle_strings.append("options")
            self.event.set()

    def _on_playlist_changed(self, playlists, playlist_id):
        """
            Add playlist to idle if mpd
            @param playlists as Playlists
            @param playlist_id as int
        """
        if playlist_id == Type.MPD:
            if not Lp().player.is_party():
                self.init_player_playlist()
                self.playlist_version += 1
                self.playlist[self.playlist_version] = []
                for track_id in Lp().playlists.get_tracks_ids(Type.MPD):
                    self.playlist[self.playlist_version].append(track_id)
                if "playlist" in self.idle_wanted_strings:
                    self.idle_strings.append("playlist")
                    self.event.set()
        elif "stored_playlist" in self.idle_wanted_strings:
            self.idle_strings.append("stored_playlist")
            self.event.set()
        # Clean history
        if len(self.playlist) > 50:
            for i in range(1, 25):
                self._playlist.pop(0)


class MpdServerDaemon(MpdServer):
    """
        Create a deamonized MPD server
        @param eth as string
        @param port as int
    """
    def __init__(self, eth="", port=6600):
        """
            Init daemon
        """
        MpdServer.__init__(self, eth, port)
        self.running = True
        event = threading.Event()
        self.thread = threading.Thread(target=self.run,
                                       args=(event,))
        self.thread.setDaemon(True)
        self.thread.start()

    def quit(self):
        """
            Stop MPD server deamon
        """
        self.running = False
        self.shutdown()
        self.server_close()



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from random import shuffle

from lollypop.define import App, StorageType
from lollypop.logger import Logger


class LocalSimilars:
    """
        Search similar artists locally
    """
    def __init__(self):
        """
            Init provider
        """
        pass

    def get_similar_artists(self, artist_names, cancellable):
        """
            Get similar artists
            @param artist_ids as [int]
            @param cancellable as Gio.Cancellable
            @return [(str, None)]
        """
        artist_ids = []
        for artist_name in artist_names:
            artist_ids.append(App().artists.get_id(artist_name)[0])
        genre_ids = App().artists.get_genre_ids(artist_ids,
                                                StorageType.COLLECTION)
        artists = App().artists.get(genre_ids, StorageType.COLLECTION)
        shuffle(artists)
        result = [(name, None) for (artist_id, name, sortname) in artists]
        if result:
            Logger.info("Found similar artists with LocalSimilars")
        return result



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib


class ButtonProgressBar(Gtk.ProgressBar):
    """
        Simple progress bar with width contraint
    """

    def __init__(self):
        Gtk.ProgressBar.__init__(self)
        self.set_property("valign", Gtk.Align.END)
        self.get_style_context().add_class("progressbar-button")

    def do_get_preferred_width(self):
        return (24, 24)


class ProgressBar(Gtk.ProgressBar):
    """
        A smart/smooth FIFO progress bar
        Many objects can register and must call set_fraction(1) to unregister
    """

    def __init__(self):
        """
            Init progress bar
        """
        Gtk.ProgressBar.__init__(self)
        self.__callers = []
        self.__fraction = 0.0
        self.__pulse_timeout = None
        self.__progress_running = False

    def add(self, caller):
        """
            Add caller
            @param caller as Instance
        """
        if caller not in self.__callers:
            self.__callers.insert(0, caller)

    def set_fraction(self, fraction, caller):
        """
            Set fraction if caller is on top.
            @param fraction as float
            @param caller as object
        """
        if not self.__callers:
            return
        if self.__pulse_timeout is not None:
            GLib.source_remove(self.__pulse_timeout)
            self.__pulse_timeout = None
        if caller == self.__callers[0]:
            self.show()
            self.__fraction = fraction
            if fraction == 0:
                Gtk.ProgressBar.set_fraction(self, 0.0)
            elif not self.__progress_running:
                self.__progress_running = True
                self.__progress_update(caller)

    def pulse(self, pulse):
        """
            Make progress bar visible/pulse if pulse is True
            @param pulse as bool
        """
        # Only pulse if nobody is using the progressbar
        if self.__callers:
            return
        if pulse:
            self.show()
            if self.__pulse_timeout is None:
                self.__pulse_timeout = GLib.timeout_add(500, self.__pulse)
        else:
            self.hide()
            if self.__pulse_timeout is not None:
                GLib.source_remove(self.__pulse_timeout)
                self.__pulse_timeout = None

#######################
# PRIVATE             #
#######################
    def __reset(self, caller):
        """
            Reset and hide progressbar
            @param caller as object
        """
        self.hide()
        self.__fraction = 0.0
        Gtk.ProgressBar.set_fraction(self, 0.0)
        self.__progress_running = False
        self.__callers.remove(caller)

    def __pulse(self):
        """
            Make progress bar pulse while visible
            @param pulse as bool
        """
        if self.is_visible():
            Gtk.ProgressBar.pulse(self)
            return True
        else:
            Gtk.ProgressBar.set_fraction(self, 0.0)
            self.__pulse_timeout = None
            return False

    def __progress_update(self, caller):
        """
            Update progressbar smoothly
            @param caller as object
        """
        if caller != self.__callers[0]:
            self.__progress_running = False
            return
        current = self.get_fraction()
        if self.__fraction < 1:
            progress = (self.__fraction - current) / 10
        else:
            progress = 0.01
        if current < self.__fraction:
            Gtk.ProgressBar.set_fraction(self, current + progress)
        if current < 1.0:
            GLib.timeout_add(10, self.__progress_update, caller)
        else:
            GLib.timeout_add(1000, self.__reset, caller)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, Gio, GLib
from gi.repository.Gio import FILE_ATTRIBUTE_STANDARD_CONTENT_TYPE

import json
from gettext import gettext as _

from lollypop.define import App, MARGIN_MEDIUM, MARGIN
from lollypop.logger import Logger


class AppListBoxRow(Gtk.ListBoxRow):
    """
        An application row
    """

    def __init__(self, app):
        """
            Init row
            @param app as Gio.AppInfo
        """
        Gtk.ListBoxRow.__init__(self)
        self.__app = app
        label = Gtk.Label.new(app.get_name())
        label.show()
        artwork = Gtk.Image.new_from_gicon(app.get_icon(), Gtk.IconSize.DIALOG)
        artwork.show()
        grid = Gtk.Grid.new()
        grid.show()
        grid.set_column_spacing(MARGIN)
        grid.add(artwork)
        grid.add(label)
        self.add(grid)

    @property
    def app(self):
        """
            Get app
            @return Gio.AppInfo
        """
        return self.__app


class AppsDialog(Gtk.Dialog):
    """
        Dialog showing apps for file
    """

    def __init__(self, f):
        """
            Init dialog
            @param f as Gio.File
        """
        Gtk.Dialog.__init__(self)
        self.__commandline = None
        self.__open_with = {}
        self.set_size_request(300, 300)
        self.set_transient_for(App().window)
        cancel_button = Gtk.Button.new_with_label(_("Cancel"))
        cancel_button.show()
        cancel_button.get_style_context().add_class("text-label")
        select_button = Gtk.Button.new_with_label(_("Select"))
        select_button.show()
        select_button.get_style_context().add_class("text-label")
        select_button.get_style_context().add_class("suggested-action")
        headerbar = Gtk.HeaderBar.new()
        headerbar.show()
        headerbar.pack_start(cancel_button)
        headerbar.pack_end(select_button)
        headerbar.set_title(_("Select an application"))
        self.set_titlebar(headerbar)
        self.__listbox = Gtk.ListBox.new()
        self.__listbox.show()
        self.__listbox.set_activate_on_single_click(False)
        self.__listbox.connect("row-selected", self.__on_row_selected)
        self.__listbox.set_sort_func(self.__sort)
        scrolled = Gtk.ScrolledWindow.new()
        scrolled.show()
        scrolled.add(self.__listbox)
        scrolled.set_property("expand", True)
        scrolled.set_shadow_type(Gtk.ShadowType.IN)
        self.get_content_area().set_property("margin", MARGIN_MEDIUM)
        self.get_content_area().add(scrolled)
        cancel_button.connect("clicked",
                              lambda x: self.response(Gtk.ResponseType.CLOSE))
        select_button.connect("clicked",
                              lambda x: self.response(Gtk.ResponseType.OK))
        self.__listbox.connect("row-activated",
                               lambda x, y: self.response(Gtk.ResponseType.OK))

        try:
            info = f.query_info("%s" % FILE_ATTRIBUTE_STANDARD_CONTENT_TYPE,
                                Gio.FileQueryInfoFlags.NONE,
                                None)
            content_type = info.get_content_type()
            self.__last_app = self.__get_last_app(content_type)
            self.connect("response", self.__on_response, content_type)
            # Set apps
            for app in Gio.AppInfo.get_all_for_type(content_type):
                if not app.get_commandline().startswith("lollypop"):
                    row = AppListBoxRow(app)
                    row.show()
                    self.__listbox.add(row)
        except Exception as e:
            Logger.error("AppsDialog::__init__(): %s", e)

    @property
    def commandline(self):
        """
            Get current commandline for dialog
            @return str/None
        """
        return self.__commandline

#######################
# PRIVATE             #
#######################
    def __get_last_app(self, content_type):
        """
            Get last application for content_type
            @param content_type as str
            @return str/None
        """
        try:
            json_data = App().settings.get_value("open-with").get_string()
            if json_data:
                self.__open_with = json.loads(json_data)
                return self.__open_with[content_type]
        except Exception as e:
            Logger.error("AppsDialog::__get_last_app(): %s", e)
            App().settings.set_value("open-with", GLib.Variant("s", ""))
        return None

    def __sort(self, row_a, row_b):
        """
            Sort function to always show last selected item
            @param row_a as AppListBoxRow
            @param row_b as AppListBoxRow
            @return bool
        """
        if row_b.app.get_commandline() == self.__last_app:
            return True
        return False

    def __on_response(self, dialog, response_id, content_type):
        """
            Save currently selected item
            @param dialog as Gtk.Dialog
            @param response_id as Gtk.ResponseType
            @param content_type as str
        """
        if response_id == Gtk.ResponseType.OK:
            if dialog.commandline is not None:
                self.__open_with[content_type] = dialog.commandline
                data = json.dumps(self.__open_with)
                App().settings.set_value("open-with",
                                         GLib.Variant("s", data))

    def __on_row_selected(self, listbox, row):
        """
            Set current commandline
            @param listbox as Gtk.ListBox
            @param row as AppListBoxRow
        """
        if row is None:
            self.__commandline = None
        else:
            self.__commandline = row.app.get_commandline()



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# Copyright (c)      2020 CodedOre
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GObject, GLib

from gettext import gettext as _

from lollypop.utils import emit_signal
from lollypop.widgets_artwork_artist import ArtistArtworkSearchWidget
from lollypop.widgets_artwork_album import AlbumArtworkSearchWidget


class ArtistArtworkMenu(Gtk.Bin):
    """
        A popover to change artwork
    """

    __gsignals__ = {
        "hidden": (GObject.SignalFlags.RUN_FIRST, None, (bool,)),
    }

    def __init__(self, artist_id, view_type, is_submenu):
        """
            Init popover
            @param artist_id as int
            @param view_type as ViewType
            @param is_submenu as bool
        """
        Gtk.Bin.__init__(self)
        self.__artist_id = artist_id
        self.view_type = view_type
        self.__is_submenu = is_submenu
        self.connect("map", self.__on_map)

    def __on_map(self, widget):
        self.__artwork_search = ArtistArtworkSearchWidget(self.__artist_id,
                                                          self.view_type, True)
        self.__artwork_search.show()
        self.__artwork_search.connect("hidden", self.__close)
        GLib.timeout_add(250, self.__artwork_search.populate)
        self.add(self.__artwork_search)

    def __close(self, action, variant):
        emit_signal(self, "hidden", True)

    @property
    def section(self):
        return None

    @property
    def submenu_name(self):
        """
            Get submenu name
            @return str
        """
        if self.__is_submenu:
            return _("Change Artwork")
        else:
            return None


class AlbumArtworkMenu(Gtk.Bin):
    """
        A popover to change artwork
    """

    __gsignals__ = {
        "hidden": (GObject.SignalFlags.RUN_FIRST, None, (bool,)),
    }

    def __init__(self, artist_id, view_type, is_submenu):
        """
            Init popover
            @param artist_id as int
        """
        Gtk.Bin.__init__(self)
        self.__artist_id = artist_id
        self.view_type = view_type
        self.__is_submenu = is_submenu
        self.connect("map", self.__on_map)

    def __on_map(self, widget):
        self.__artwork_search = AlbumArtworkSearchWidget(self.__artist_id,
                                                         self.view_type, True)
        self.__artwork_search.show()
        self.__artwork_search.connect("hidden", self.__close)
        GLib.timeout_add(250, self.__artwork_search.populate)
        self.add(self.__artwork_search)

    def __close(self, action, variant):
        emit_signal(self, "hidden", True)

    @property
    def section(self):
        return None

    @property
    def submenu_name(self):
        """
            Get submenu name
            @return str
        """
        if self.__is_submenu:
            return _("Change Artwork")
        else:
            return None



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import GLib, Gio

import json

from lollypop.define import App, GOOGLE_API_ID
from lollypop.utils import get_network_available, emit_signal
from lollypop.logger import Logger


class ArtworkDownloader:
    """
        Download artwork from the web
        Need to be herited by an ArtworkManager
    """

    def __init__(self):
        """
            Init art downloader
        """
        self.__cancellable = Gio.Cancellable()

    def search_artwork_from_google(self, search, cancellable):
        """
            Get google uri for search
            @param search as str
            @param cancellable as Gio.Cancellable
        """
        if not get_network_available("GOOGLE"):
            emit_signal(self, "uri-artwork-found", None)
            return
        key = App().settings.get_value("cs-api-key").get_string() or\
            App().settings.get_default_value("cs-api-key").get_string()
        uri = "https://www.googleapis.com/" +\
              "customsearch/v1?key=%s&cx=%s" % (key, GOOGLE_API_ID) +\
              "&q=%s&searchType=image" % GLib.uri_escape_string(search,
                                                                "",
                                                                False)
        App().task_helper.load_uri_content(uri,
                                           cancellable,
                                           self.__on_load_google_content)

    def search_artwork_from_startpage(self, search, cancellable):
        """
            Get google uri for search
            @param search as str
            @param cancellable as Gio.Cancellable
        """
        if not get_network_available("STARTPAGE"):
            emit_signal(self, "uri-artwork-found", None)
            return
        uri = "https://www.startpage.com/do/search?flimgsize=isz%3Al"
        uri += "&image-size-select=&flimgexwidth=&flimgexheight=&abp=-1"
        uri += "&cat=pics&query=%s" % GLib.uri_escape_string(search, "", False)
        App().task_helper.load_uri_content(uri,
                                           cancellable,
                                           self.__on_load_startpage_content)

    @property
    def cancellable(self):
        """
            Get global cancellable
            @return Gio.Cancellable
        """
        return self.__cancellable

#######################
# PROTECTED           #
#######################
    def _get_musicbrainz_mbid(self, mbid_type, string, cancellable):
        """
            Get musicbrainz mbid for type and string
            @param mbid_type as str ("artist" or "album")
            @param string as str
            @param cancellable as Gio.Cancellable
            @return str
        """
        try:
            if mbid_type == "artist":
                uri = "http://musicbrainz.org/ws/2/artist/" +\
                      "?query=%s&fmt=json"
            else:
                uri = "http://musicbrainz.org/ws/2/release-group/" +\
                      "?query=%s&fmt=json"
            string = GLib.uri_escape_string(string, None, True)
            (status, data) = App().task_helper.load_uri_content_sync(
                                                          uri % string,
                                                          cancellable)
            if status:
                decode = json.loads(data.decode("utf-8"))
                if mbid_type == "artist":
                    for item in decode["artists"]:
                        return item["id"]
                else:
                    mbid = None
                    for item in decode["release-groups"]:
                        if "primary-type" not in item.keys():
                            continue
                        if item["primary-type"] == "Album":
                            mbid = item["id"]
                            break
                        elif item["primary-type"] == "EP" and mbid is None:
                            mbid = item["id"]
                return mbid
        except:
            Logger.warning("MusicBrainz: %s", uri)
        return None

#######################
# PRIVATE             #
#######################
    def __on_load_google_content(self, uri, loaded, content):
        """
            Extract uris from content
            @param uri as str
            @param loaded as bool
            @param content as bytes
        """
        try:
            if not loaded:
                emit_signal(self, "uri-artwork-found", None)
                return
            decode = json.loads(content.decode("utf-8"))
            results = []
            for item in decode["items"]:
                if item["link"] is not None:
                    results.append((item["link"], "Google"))
            emit_signal(self, "uri-artwork-found", results)
        except Exception as e:
            emit_signal(self, "uri-artwork-found", None)
            Logger.error(
                "ArtworkDownloader::__on_load_google_content(): %s: %s"
                % (e, content))

    def __on_load_startpage_content(self, uri, loaded, content):
        """
            Extract uris from content
            @param uri as str
            @param loaded as bool
            @param content as bytes
        """
        import re

        def search_in_data(lines, found_uris=[]):
            if lines:
                line = lines.pop(0)
                # Do not call findall if nothing to find
                if line.find("oiu=") != -1:
                    res = re.findall(r'.*oiu=([^&]*).*', line)
                    for data in res:
                        uri = GLib.uri_unescape_string(data, "")
                        if uri in found_uris or uri is None:
                            continue
                        found_uris.append(uri)
                GLib.idle_add(search_in_data, lines, found_uris)
            else:
                results = [(uri, "Startpage") for uri in found_uris]
                emit_signal(self, "uri-artwork-found", results)

        try:
            if not loaded:
                emit_signal(self, "uri-artwork-found", None)
                return
            lines = content.decode("utf-8").splitlines()
            search_in_data(lines)
        except Exception as e:
            emit_signal(self, "uri-artwork-found", None)
            Logger.error("ArtworkDownloader::__on_load_startpage_content(): %s"
                         % e)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib, GObject, Handy

from lollypop.define import App, Type
from lollypop.utils import emit_signal
from lollypop.view import View
from lollypop.container_stack import StackContainer
from lollypop.container_notification import NotificationContainer
from lollypop.container_scanner import ScannerContainer
from lollypop.container_playlists import PlaylistsContainer
from lollypop.container_lists import ListsContainer
from lollypop.container_views import ViewsContainer
from lollypop.container_filter import FilterContainer
from lollypop.progressbar import ProgressBar


class Grid(Gtk.Grid):
    def do_get_preferred_width(self):
        return 200, 700


class Container(Gtk.Overlay, NotificationContainer,
                ScannerContainer, PlaylistsContainer,
                ListsContainer, ViewsContainer, FilterContainer):
    """
        Main view management
    """

    __gsignals__ = {
        "can-go-back-changed": (GObject.SignalFlags.RUN_FIRST, None, (bool,)),
    }

    def __init__(self):
        """
            Init container
            @param view_type as ViewType, will be appended to any created view
        """
        Gtk.Overlay.__init__(self)
        NotificationContainer.__init__(self)
        ScannerContainer.__init__(self)
        PlaylistsContainer.__init__(self)
        ViewsContainer.__init__(self)

    def setup(self):
        """
            Setup container
        """
        self.__widget = Handy.Leaflet()
        self.__widget.show()
        self.__widget.connect("notify::folded", self.__on_folded)
        self.__sub_widget = Handy.Leaflet()
        self.__sub_widget.show()
        self.__focused_view = None
        self._stack = StackContainer()
        self._stack.show()
        ListsContainer.__init__(self)
        self.__progress = ProgressBar()
        self.__progress.get_style_context().add_class("progress-bottom")
        self.__progress.set_property("valign", Gtk.Align.END)
        self.add_overlay(self.__progress)
        search_action = App().lookup_action("search")
        search_action.connect("activate", self.__on_search_activate)
        self.__widget.add(self.sidebar)
        self.__grid_view = Grid()
        self.__grid_view.set_orientation(Gtk.Orientation.VERTICAL)
        self.__grid_view.show()
        self.__sub_widget.add(self.left_list)
        self.__sub_widget.add(self.__grid_view)
        self.__grid_view.attach(self._stack, 0, 0, 1, 1)
        self.__widget.add(self.__sub_widget)
        self.__widget.set_visible_child(self.__sub_widget)
        self.__sub_widget.set_visible_child(self.__grid_view)
        self.add(self.__widget)
        FilterContainer.__init__(self)

    def stop(self):
        """
            Stop current view from processing
        """
        view = self._stack.get_visible_child()
        if view is not None:
            view.stop()

    def reload_view(self):
        """
            Reload current view
        """
        view = self._stack.get_visible_child()
        if view is not None and view.args is not None:
            cls = view.__class__
            new_view = cls(**view.args)
            new_view.populate()
            new_view.show()
            self._stack.add(new_view)
            self._stack.set_visible_child(new_view)
        else:
            App().lookup_action("reload").change_state(GLib.Variant("b", True))

    def set_focused_view(self, view):
        """
            Set focused view
            @param view as View
        """
        self.__focused_view = view

    def go_home(self):
        """
            Go back to first page
        """
        self.__widget.set_visible_child(self.sidebar)
        self._stack.clear()
        emit_signal(self, "can-go-back-changed", False)

    def go_back(self):
        """
            Go back in history
        """
        if self._stack.history.count > 0:
            self._stack.go_back()
            self.set_focused_view(self._stack.get_visible_child())
        elif self.__sub_widget.get_folded() and\
                self.__sub_widget.get_visible_child() != self.left_list:
            self.__sub_widget.set_visible_child(self.left_list)
            self._stack.clear()
        elif self.__widget.get_folded() and\
                self.__widget.get_visible_child() != self.sidebar:
            self.__widget.set_visible_child(self.sidebar)
            self._stack.clear()
        emit_signal(self, "can-go-back-changed", self.can_go_back)

    @property
    def can_go_back(self):
        """
            True if can go back
            @return bool
        """
        if self.__widget.get_folded():
            return self.__widget.get_visible_child() != self._sidebar
        else:
            return self._stack.history.count > 0

    @property
    def widget(self):
        """
            Get main widget
            @return Handy.Leaflet
        """
        return self.__widget

    @property
    def sub_widget(self):
        """
            Get sub widget
            @return Handy.Leaflet
        """
        return self.__sub_widget

    @property
    def grid_view(self):
        """
            Get grid view
            @return Gtk.Grid
        """
        return self.__grid_view

    @property
    def focused_view(self):
        """
            Get focused view
            @return View
        """
        return self.__focused_view

    @property
    def view(self):
        """
            Get current view
            @return View
        """
        view = self._stack.get_visible_child()
        if view is not None and isinstance(view, View):
            return view
        return None

    @property
    def stack(self):
        """
            Container stack
            @return stack as Gtk.Stack
        """
        return self._stack

    @property
    def progress(self):
        """
            Progress bar
            @return ProgressBar
        """
        return self.__progress

############
# PRIVATE  #
############
    def __on_folded(self, *ignore):
        """
            Reload main view if needed
        """
        if not App().window.folded and self.view is None:
            self.show_view(self.sidebar.selected_ids)

    def __on_search_activate(self, action, variant):
        """
            @param action as Gio.SimpleAction
            @param variant as GLib.Variant
        """
        if App().window.folded:
            search = variant.get_string()
            App().window.container.show_view([Type.SEARCH], search)



# Copyright (c) 2014-2021 Cedric Bellegarde <cedric.bellegarde@adishatz.org>
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
# You should have received a copy of the GNU General Public License
# along with this program. If not, see <http://www.gnu.org/licenses/>.

from gi.repository import Gtk, GLib, Pango

from gettext import gettext as _

from lollypop.view import View
from lollypop.utils import get_network_available
from lollypop.define import ViewType, StorageType, Size, App
from lollypop.view_albums_line import AlbumsPopularsLineView
from lollypop.view_albums_line import AlbumsRandomGenresLineView
from lollypop.view_artists_line import ArtistsRandomLineView
from lollypop.widgets_banner_today import TodayBannerWidget
from lollypop.helper_signals import signals_map


class SuggestionsView(View):
    """
        View showing suggestions to user
    """

    @signals_map
    def __init__(self, storage_type, view_type):
        """
            Init view
            @param storage_type as StorageType
            @param view_type as ViewType
        """
        View.__init__(self, storage_type, view_type | ViewType.OVERLAY)
        self.__grid = Gtk.Grid()
        self.__grid.get_style_context().add_class("padding")
        self.__grid.set_row_spacing(5)
        self.__grid.set_orientation(Gtk.Orientation.VERTICAL)
        self.__grid.show()
        if App().tracks.count() == 0:
            self.__welcome_screen()
            self.add_widget(self.__grid, None)
        else:
            album = TodayBannerWidget.get_today_album()
            if album is not None:
                self.__banner = TodayBannerWidget(album, self.view_type)
                self.__banner.show()
            else:
                self.__banner = None
            self.add_widget(self.__grid, self.__banner)
        return [
            (App().settings, "changed::suggestions-mask", "_on_mask_changed"),
        ]

    def populate(self):
        """
            Populate view
        """
        for cls in [AlbumsPopularsLineView,
                    ArtistsRandomLineView,
                    AlbumsRandomGenresLineView]:
            view = cls(self.storage_type, self.view_type)
            view.populate()
            self.__grid.add(view)
        self.__add_web_views()

    @property
    def filtered(self):
        """
            Get filtered widgets
            @return [Gtk.Widget]
        """
        children = []
        for child in reversed(self.__grid.get_children()):
            children += child.children
        return children

    @property
    def args(self):
        """
            Get default args for __class__
            @return {}
        """
        return {"storage_type": self.storage_type,
                "view_type": self.view_type}

#######################
# PROTECTED           #
#######################
    def _on_mask_changed(self, settings, value):
        """
            Reload web views
            @param settings as Gio.Settings
            @param value as str
        """
        self.__add_web_views()

    def _on_map(self, widget):
        """
            Set initial view state
            @param widget as GtK.Widget
        """
        View._on_map(self, widget)
        if self.view_type & ViewType.SCROLLED:
            self.scrolled.grab_focus()

#######################
# PRIVATE             #
#######################
    def __add_web_views(self):
        """
            Add web views
        """
        from lollypop.view_albums_line import AlbumsStorageTypeLineView
        for child in self.__grid.get_children():
            if isinstance(child, AlbumsStorageTypeLineView):
                child.destroy()

        mask = App().settings.get_value("suggestions-mask").get_int32()
        if get_network_available("YOUTUBE"):
            for (title, storage_type) in [
                   (_("Suggestions from Spotify"),
                    StorageType.SPOTIFY_SIMILARS),
                   (_("New releases on Spotify"),
                    StorageType.SPOTIFY_NEW_RELEASES),
                   (_("Top albums on Deezer"),
                    StorageType.DEEZER_CHARTS)]:
                if not storage_type & mask:
                    continue
                view = AlbumsStorageTypeLineView(title,
                                                 self.storage_type,
                                                 self.view_type)
                view.populate(storage_type)
                self.__grid.add(view)

    def __welcome_screen(self):
        """
            Show welcome screen if view empty
        """
        label = Gtk.Label.new(_("Welcome to Lollypop"))
        label.get_style_context().add_class("text-xx-large")
        label.set_property("valign", Gtk.Align.END)
        label.set_ellipsize(Pango.EllipsizeMode.END)
        label.set_vexpand(True)
        label.show()
        label.get_style_context().add_class("opacity-transition")
        image = Gtk.Image.new_from_icon_name("org.gnome.Lollypop",
                                             Gtk.IconSize.INVALID)
        image.set_pixel_size(Size.MINI)
        image.show()
        image.get_style_context().add_class("image-rotate-fast")
        image.get_style_context().add_class("opacity-transition")
        image.set_hexpand(True)
        image.set_vexpand(True)
        image.set_property("valign", Gtk.Align.START)
        self.__grid.add(label)
        self.__grid.add(image)
        GLib.idle_add(label.set_state_flags, Gtk.StateFlags.VISITED, True)
        GLib.idle_add(image.set_state_flags, Gtk.StateFlags.VISITED, True)
